import protein_sol_api
import os
import argparse
import glob

from Bio.PDB import PDBParser
from Bio.SeqUtils import seq1
from Bio import SeqIO

from typing import Tuple, List

import pandas as pd
import numpy as np

pdb_parser: PDBParser = PDBParser(QUIET=True)
AMINO_ACIDS = list("ACDEFGHIKLMNPQRSTVWY")
PROTEINSOL_MIN_LENGTH = 21


def get_sequences_from_pdb_file(file_path: str, chains: List[str]) -> list[tuple[str, str]]:
    """
    Get pairs of sequences and chain identifiers from the pdb structure file.
    """
    with open(file_path) as f:
        parsed_structure = pdb_parser.get_structure("", f)
        available_chains = [c.get_id() for c in parsed_structure[0].get_chains()]
        seqs_concat = []
        for chain in chains:
            if chain not in available_chains:
                raise ValueError(
                    f"Chain {chain} not found in structure {os.path.basename(file_path)}, available chains: {available_chains}"
                )
            sequence = []
            for residue in parsed_structure[0][chain]:
                if residue.has_id("CA"):
                    sequence.append(seq1(residue.get_resname()))
            seqs_concat.append(("".join(sequence), chain))
        return seqs_concat


def filter_short_sequences(sequences: dict, min_length: int = 21) -> Tuple[dict, dict]:
    """
    Split the sequences into two dictionaries: one with sequences longer than min_length and one with sequences shorter than min_length.
    """
    min_length_sequences = {}
    short_sequences = {}
    for seq_id, sequence in sequences.items():
        if len(sequence) >= min_length:
            min_length_sequences[seq_id] = sequence
        else:
            short_sequences[seq_id] = sequence
    return short_sequences, min_length_sequences


def get_charge_perc(x) -> Tuple[float, float]:
    """
    Calculate the percentage of positive and negative charges in a list of charges.
    """
    x_wo_none = [value for value in x if value is not None]
    positive_perc = len([value for value in x_wo_none if value > 0]) / len(x_wo_none) * 100
    negative_perc = len([value for value in x_wo_none if value < 0]) / len(x_wo_none) * 100
    return (positive_perc, negative_perc)


def charged_avg(x) -> Tuple[float, float]:
    """
    Calculate the average positive and negative charges in a list of charges.
    """
    x_wo_none = [value for value in x if value is not None]
    positive_avg = np.mean([value for value in x_wo_none if value > 0])
    negative_avg = np.mean([value for value in x_wo_none if value < 0])
    return (positive_avg, negative_avg)


def expand_chains_per_structure(filename: str, seqs_concat: list[tuple[str, str]]) -> dict[str, str]:
    """
    Create a dictionary mapping of structure with chain ID to its corresponding sequences.

    Example:
        >>> expand_chains_per_structure("1abc", [("ACDE", "A"), ("FGHI", "B")])
        {'1abc_A': 'ACDE', '1abc_B': 'FGHI'}
    """
    if not seqs_concat:
        return {}
    return {"_".join([filename, chain]): seq for seq, chain in seqs_concat}


def weighted_avg(x: pd.DataFrame, col: str) -> float:
    """
    Computes weighted average for a given column with respect to sequence length.
    """
    return (x[col] * x["seq_len"]).sum() / x["seq_len"].sum()


def custom_agg(x: pd.DataFrame, agg_map: dict, weighted_cols: list) -> pd.Series:
    """
    Aggregate columns with specified functions and do weighted averages on selected columns.
    """
    result = {}
    for col, func in agg_map.items():
        result[col] = func(x[col])
    for col in weighted_cols:
        result[col] = weighted_avg(x, col)
    return pd.Series(result)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("input_path", type=str)
    parser.add_argument("output_csv", type=str)
    parser.add_argument("-c", "--chains", type=str, default="A", help="Chain to extract the sequence from")
    options = parser.parse_args()

    chains = [chain.strip() for chain in options.chains.split(",")]

    if os.path.isdir(options.input_path):
        paths = sorted(glob.glob(os.path.join(options.input_path, "*.pdb")))
        print(f"Reading sequences from {len(paths):,} PDBs")
        sequences_by_id = {
            struct_chain_id: sequence
            for path in paths
            for struct_chain_id, sequence in expand_chains_per_structure(
                os.path.splitext(os.path.basename(path))[0], get_sequences_from_pdb_file(path, chains=chains)
            ).items()
        }
    elif options.input_path.endswith((".pdb",)):
        sequences_by_id = {
            struct_chain_id: sequence
            for struct_chain_id, sequence in expand_chains_per_structure(
                os.path.splitext(os.path.basename(options.input_path))[0],
                get_sequences_from_pdb_file(options.input_path, chains=chains),
            ).items()
        }
    elif options.input_path.endswith((".fasta", ".fa")):
        sequences_by_id = {
            record.id.split("|")[0]: str(record.seq) for record in SeqIO.parse(options.input_path, "fasta")
        }
    else:
        raise ValueError("Input must be a directory with PDB files, a PDB file, or a FASTA file")

    # ProteinSol only predicts for sequences >= 21aa
    # Filter out too short sequences
    short_sequences_by_id, normal_sequences_by_id = filter_short_sequences(sequences_by_id, PROTEINSOL_MIN_LENGTH)

    # If no long-enough sequences are found
    if not normal_sequences_by_id:
        with open(options.output_csv, "w") as f:
            df = pd.DataFrame({"id": short_sequences_by_id.keys()})
            df["error"] = "Sequence too short"
    else:
        result = protein_sol_api.protein_sol(list(normal_sequences_by_id.values()))
        df = pd.DataFrame(result)
        df["seq_len"] = pd.Series(list(normal_sequences_by_id.values())).apply(lambda x: x.__len__())

        # Add ids to the dataframe
        df.insert(0, "id", normal_sequences_by_id.keys())

        # Add short sequences to the dataframe
        short_df = pd.DataFrame({"id": short_sequences_by_id.keys()})
        short_df["error"] = "Sequence too short"
        df = pd.concat([df, short_df], ignore_index=True)

    # Separate structure id from processed chains
    df["chains"] = df["id"].apply(lambda x: x.rsplit("_", 1)[1] if "_" in x else "")
    df["id"] = df["id"].apply(lambda x: x.rsplit("_", 1)[0] if "_" in x else x)

    # Aggregate protein id, chains, errors and compute weighted average for specified numeric values
    AGG_MAP = {
        "id": lambda x: np.unique(x)[0],
        "chains": lambda x: ",".join(x.dropna().astype(str)),
        "error": lambda x: list(x.dropna().astype(str)),
    }

    NUMERIC_VALUES = ["percent-sol", "scaled-sol", "population-sol"]
    df = df.groupby("id").apply(lambda g: custom_agg(g, AGG_MAP, NUMERIC_VALUES)).reset_index(drop=True)
    df.to_csv(options.output_csv, index=False)



from copy import deepcopy
from typing import Callable

import streamlit as st

from ovo.app.utils.cached_db import get_cached_workflow_pools_and_jobs
from datetime import datetime
import timeago
from ovo.core.database import Pool, WorkflowTypes


def history_dropdown_component(
    page_key: str, workflow_name: str, filter_func: Callable = None, include_subclasses: bool = True
):
    if page_key not in st.session_state.initial_workflows:
        raise ValueError(
            "history_dropdown_component should be called after initialize_workflow() so that st.session_state.initial_workflows is set"
        )

    # Get list of workflow names including subclasses
    if include_subclasses:
        workflow_names = WorkflowTypes.get_subclass_names(workflow_name)
    else:
        workflow_names = [workflow_name]

    pools_by_id, jobs_by_id = get_cached_workflow_pools_and_jobs(st.session_state.project.id, workflow_names)
    if filter_func:
        pools_by_id = {
            pool_id: pool
            for pool_id, pool in pools_by_id.items()
            if filter_func(jobs_by_id[pool.design_job_id].workflow)
        }
    num_options = len(pools_by_id)

    previous_pool_id = st.session_state.get("history_pool_id")
    pool_ids_by_date = [p.id for p in sorted(pools_by_id.values(), key=lambda p: p.created_date_utc, reverse=True)]
    pool_id: str = st.selectbox(
        "Re-use previous run:",
        options=pool_ids_by_date,
        format_func=lambda pool_id: format_label(pool_id, pools_by_id),
        index=None,
        key="history_dropdown",
        placeholder=("1 run available" if num_options == 1 else f"{num_options} runs available")
        if num_options
        else "No runs in this project",
        help="Select a previous run to load its parameters. "
        "If you clear the selection, the parameters will be cleared and history reloaded.",
    )

    if pool_id:
        pool = pools_by_id[pool_id]
        workflow = jobs_by_id[pool.design_job_id].workflow

    if pool_id != previous_pool_id:
        if pool_id:
            workflow.preview_job_id = None
            st.session_state.workflows[page_key] = workflow
            st.session_state.pool_inputs[page_key] = None, pool.name, pool.description
        else:
            # Reset to initial workflow when selection is cleared
            st.session_state.workflows[page_key] = deepcopy(st.session_state.initial_workflows[page_key])

        st.session_state["history_pool_id"] = pool_id
        st.rerun()


def format_label(pool_id: str, pools_by_id: dict[str, Pool]):
    delta = datetime.utcnow() - pools_by_id[pool_id].created_date_utc
    return "{id} | {name} | {when_created}".format(
        id=pool_id,
        name=pools_by_id[pool_id].name,
        # NOTE: we need to use "a moment ago" for times less than 1 hour
        # to avoid the streamlit selectbox being re-created and discarding user's selection
        when_created=timeago.format(delta) if delta.total_seconds() > 3600 else "a moment ago",
    )



from typing import Dict, List

import pandas as pd
import streamlit as st

from ovo import db, storage
from ovo.app.components.descriptor_job_components import refresh_descriptors
from ovo.app.components.descriptor_table import residue_number_descriptor_detail_table
from ovo.app.components.descriptor_tiles import get_residue_presence_df
from ovo.app.components.download_component import download_job_designs_component
from ovo.app.components.molstar_custom_component import molstar_custom_component, StructureVisualization
from ovo.app.components.navigation import design_navigation_selector
from ovo.app.utils.cached_db import (
    get_cached_pools,
    get_cached_design,
    get_cached_design_jobs,
    get_cached_design_job,
    get_cached_available_descriptors,
    get_cached_pool,
)
from ovo.core.database import Design, DesignWorkflow, NumericDescriptor, Descriptor
from ovo.core.database.descriptors_refolding import AF2_PRIMARY_IPAE, AF2_PRIMARY_IPTM
from ovo.core.database.descriptors_rfdiffusion import (
    INTERFACE_TARGET_RESIDUES,
    N_CONTACTS_TO_INTERFACE,
    N_CONTACTS_TO_HOTSPOTS,
    N_HOTSPOTS_ON_INTERFACE,
    PYROSETTA_DDG,
    PYROSETTA_CMS,
    PYROSETTA_SAP_SCORE,
)
from ovo.core.logic.descriptor_logic import get_wide_descriptor_table

# Interface-related descriptors grouped by category
INTERFACE_DESCRIPTORS_BACKBONE = [
    N_CONTACTS_TO_INTERFACE,
    N_CONTACTS_TO_HOTSPOTS,
    N_HOTSPOTS_ON_INTERFACE,
    INTERFACE_TARGET_RESIDUES,
]

INTERFACE_DESCRIPTORS_PYROSETTA = [
    PYROSETTA_DDG,
    PYROSETTA_CMS,
    PYROSETTA_SAP_SCORE,
]

INTERFACE_DESCRIPTORS_AF2 = [
    AF2_PRIMARY_IPAE,
    AF2_PRIMARY_IPTM,
]

ALL_INTERFACE_DESCRIPTORS = INTERFACE_DESCRIPTORS_BACKBONE + INTERFACE_DESCRIPTORS_PYROSETTA + INTERFACE_DESCRIPTORS_AF2


@st.fragment
def interface_fragment(pool_ids: List[str], design_ids: List[str] | None = None):
    """Main interface analysis view."""
    pools = get_cached_pools(pool_ids)

    if design_ids is None:
        # design_ids not explicitly passed, use all accepted designs in the selected pools
        design_ids = sorted(db.select_values(Design, "id", pool_id__in=pool_ids, accepted=True))

        if not design_ids:
            st.write(
                "No accepted designs in the selected "
                + ("pools" if len(pool_ids) > 1 else "pool")
                + ". Please mark some designs as accepted in the **Jobs** page."
            )
            return

    elif not design_ids:
        # Empty list explicitly passed to design_ids
        st.write("No designs selected")
        return

    st.header(f"Interface analyzer | {len(design_ids):,} {'design' if len(design_ids) == 1 else 'designs'}")

    design_jobs = get_cached_design_jobs([pool.design_job_id for pool in pools if pool.design_job_id])
    if not design_jobs:
        st.warning("Selected pools have no design jobs. Please select pools with completed design jobs.")
        return

    badges = [f":grey-badge[{pool.name}]" for pool in pools]
    st.write("Selected pools: " + " ".join(badges))

    # Refresh descriptors
    refresh_descriptors(
        design_ids=design_ids,
    )

    descriptors_by_key = get_cached_available_descriptors(design_ids)
    interface_descriptors_by_key = {
        desc.key: desc for desc in ALL_INTERFACE_DESCRIPTORS if desc.key in descriptors_by_key
    }

    if INTERFACE_TARGET_RESIDUES.key not in interface_descriptors_by_key:
        st.warning("Interface analysis metrics are not provided by this design workflow.")
        return

    descriptors_df = get_wide_descriptor_table(
        design_ids=design_ids, descriptor_keys=interface_descriptors_by_key.keys(), nested=False, human_readable=False
    )

    # Average interface metrics
    st.write("#### Interface metrics summary")
    interface_metrics_summary(descriptors_df, interface_descriptors_by_key)

    # Summary table per target interface residue
    st.write("#### Target interface residue statistics")
    st.text(
        "Shows which target residues are in contact with the binder backbone (CA within 8Å)",
        help="Note that pools may have different target structures and residue numberings.",
    )
    # NOTE: different pools may have different target structures and thus different interface residues and residue numberings
    interface_residues_table(descriptors_df)

    # Summary table per design and per target interface residue
    st.write("#### Target interface residues by design")
    st.text(
        "Matrix showing which target residues are in contact with the binder backbone (CA within 8Å) in each design",
        help="Note that pools may have different target structures and residue numberings.",
    )
    # NOTE: different pools may have different target structures and thus different interface residues and residue numberings
    interface_residues_table_by_design(descriptors_df)

    # TODO: Target structure view
    # We could visualize the target and color all interface residues on it (darker residues = more designs with that contact)
    # Issue: In the pools, we can have different target structures, so we would need to group by target structure first...

    # Individual design exploration
    st.write("#### Design details")

    interface_design_visualization_fragment(design_ids, descriptors_df, interface_descriptors_by_key)


@st.fragment()
def interface_metrics_summary(descriptors_df: pd.DataFrame, descriptors_by_key: Dict[str, Descriptor]):
    """Display summary statistics for interface metrics.

    Args:
        descriptors_df: DataFrame with descriptor values (can be multiple designs or single design)
        descriptors_by_key: Dictionary mapping descriptor keys to descriptor objects
    """
    if descriptors_df.empty:
        st.info("No interface metrics available for the selected designs.")
        return

    # Determine if we should show averages or individual values
    is_single_design = len(descriptors_df) == 1

    # Display metrics in columns
    n_cols = 3
    cols = st.columns(n_cols)

    col_idx = 0
    for col_key in descriptors_df.columns:
        descriptor = descriptors_by_key.get(col_key)
        values = descriptors_df[col_key].dropna()

        if descriptor is None or not isinstance(descriptor, NumericDescriptor):
            continue

        if len(values) == 0:
            continue

        # Compute value: mean for multiple designs, single value otherwise
        value = values.mean() if not is_single_design else values.iloc[0]

        # Format values based on column name
        if "pTM" in col_key or "pLDDT" in col_key:
            formatted_value = f"{value:.3f}"
        elif "RMSD" in col_key or "PAE" in col_key or "ddG" in col_key:
            formatted_value = f"{value:.2f}"
        else:
            formatted_value = f"{value:.1f}"

        # Add "Average " prefix for multiple designs
        label = f"Average {descriptor.name}" if not is_single_design else descriptor.name

        with cols[col_idx % n_cols]:
            st.metric(
                label=label,
                value=formatted_value,
                delta=None,
                help=descriptor.description,
            )

        col_idx += 1

    if col_idx == 0:
        st.info("No interface metrics available.")


@st.fragment()
def interface_residues_table(descriptors_df: pd.DataFrame):
    """Display table of interface residues and their presence across designs."""
    descriptor = INTERFACE_TARGET_RESIDUES

    # Check if interface residues column exists
    if descriptor.key not in descriptors_df.columns:
        st.warning("No interface residue data available.")
        return

    # Get residue values from DataFrame
    residue_values = descriptors_df[descriptor.key].dropna().tolist()

    if not residue_values:
        st.warning("No interface residue data available.")
        return

    residue_df = get_residue_presence_df(descriptor, residue_values, len(descriptors_df))

    if residue_df is None or residue_df.empty:
        st.warning("No interface residue data available.")
    else:
        st.dataframe(
            residue_df,
            column_config={
                descriptor.name: st.column_config.Column(descriptor.name),
                "% of designs": st.column_config.ProgressColumn(
                    "% of designs",
                    format="%.1f%%",
                    min_value=0,
                    max_value=100,
                    help="Percentage of designs with contact to this residue",
                ),
            },
            hide_index=True,
            width="stretch",
            height="auto",
        )


@st.fragment()
def interface_residues_table_by_design(descriptors_df: pd.DataFrame):
    """Display table showing which interface residues are present in each design."""
    descriptor = INTERFACE_TARGET_RESIDUES

    # Check if interface residues column exists
    if descriptor.key not in descriptors_df.columns:
        st.warning("No interface residue data available.")
        return

    # Get residue values as a Series with design IDs as index
    descriptor_values = descriptors_df[descriptor.key]

    if not any(v is not None and pd.notna(v) for v in descriptor_values):
        st.warning("No interface residue data available.")
        return

    table_data, column_config, caption, format_func = residue_number_descriptor_detail_table(
        descriptor, descriptor_values
    )
    # st.caption(caption)
    st.dataframe(
        table_data,
        column_config=column_config,
        width="stretch",
        height="auto",
    )


def get_molstar_residue_selections(interface_residues: List[str]) -> List[str]:
    """Convert interface residue strings to molstar selection format."""
    selections = []
    for res in interface_residues:
        res_number = int(res[1:])
        selections.append(f"{res}-{res_number}")
    return selections


@st.fragment()
def design_interface_detail(
    design_id: str, descriptors_df: pd.DataFrame, interface_descriptors_by_key: Dict[str, Descriptor]
):
    """Display detailed interface information for a single design."""
    design = get_cached_design(design_id)

    # Get interface residues for this design from descriptors_df
    interface_residues_str = (
        descriptors_df.loc[design_id, INTERFACE_TARGET_RESIDUES.key]
        if design_id in descriptors_df.index and INTERFACE_TARGET_RESIDUES.key in descriptors_df.columns
        else None
    )

    # Parse residues list once
    interface_residues = []
    if interface_residues_str and pd.notna(interface_residues_str):
        interface_residues = [r.strip() for r in str(interface_residues_str).split(",") if r.strip()]

    # TODO: Improve visualization of the interface - Maybe it is possible to display contacts?
    colors = {
        "chain-id": "Color by chain",
        "hydrophobicity": "Color by hydrophobicity (green = hydrophobic, red = hydrophilic)",
        "residue-charge": "Color by residue charge (blue = positive, red = negative)",
    }
    color = st.selectbox(
        "Color scheme",
        options=list(colors.keys()),
        format_func=lambda x: colors[x],
        key="color_scheme_input",
        label_visibility="collapsed",
        width=500,
    )
    molstar_custom_component(
        structures=[
            StructureVisualization(
                pdb=storage.read_file_str(design.structure_path),
                color=color,
                representation_type="cartoon+ball-and-stick",
                highlighted_selections=get_molstar_residue_selections(interface_residues),
            )
        ],
        key="interface_highlighted_structure",
    )

    # Display interface residues
    st.write("**Interface Residues**")
    if interface_residues:
        badges = " ".join([f":gray-badge[{res}]" for res in interface_residues])
        st.write(badges)
    else:
        st.info("No interface residues found.")

    # Display interface metrics
    if design_id in descriptors_df.index:
        interface_metrics_summary(
            descriptors_df.loc[[design_id]],
            interface_descriptors_by_key,
        )
    else:
        st.info("No interface metrics available for this design.")

    st.markdown("#### Sequence")

    pool = get_cached_pool(design.pool_id)
    design_job = get_cached_design_job(pool.design_job_id) if pool.design_job_id else None
    WorkflowType = type(design_job.workflow) if design_job and design_job.workflow else DesignWorkflow
    WorkflowType.visualize_single_design_sequences(design_id)

    st.markdown(f"#### Download {design_id}")

    download_job_designs_component(design_ids=[design_id], pools=[pool], key="single")


@st.fragment
def interface_design_visualization_fragment(
    design_ids: List[str], descriptors_df: pd.DataFrame, interface_descriptors_by_key: dict
):
    """Fragment for browsing individual designs."""
    # Optional filter for designs with contact to specific target residue
    if INTERFACE_TARGET_RESIDUES.key not in descriptors_df.columns:
        st.warning("No interface residue data available")
        return

    interface_residues = set(
        x.strip() for row in descriptors_df[INTERFACE_TARGET_RESIDUES.key].dropna() for x in row.split(",")
    )
    target_residue = st.selectbox(
        "Filter designs by target interface residue",
        options=interface_residues,
        index=None,
        placeholder="Show all designs (no residue filter)",
        help=(
            "Select a target interface residue to only show designs where the binder contacts this residue. "
            "Residue contacts are determined by backbone CA atoms within 8Å of the target. "
        ),
    )
    if target_residue:
        selected_design_ids = descriptors_df[
            descriptors_df[INTERFACE_TARGET_RESIDUES.key].apply(
                lambda x: target_residue in str(x).split(",") if pd.notna(x) else False
            )
        ].index.tolist()
        st.write(f"Showing {len(selected_design_ids):,} designs that interact with target residue {target_residue}")
    else:
        st.write("Showing all designs. Use the dropdown above to select designs that bind a specific target residue.")
        selected_design_ids = design_ids

    # Use the navigation selector component
    labels_by_design_id = descriptors_df.loc[selected_design_ids, INTERFACE_TARGET_RESIDUES.key].to_dict()
    design_id = design_navigation_selector(
        selected_design_ids, key="interface_selected_design", fmt=labels_by_design_id
    )

    st.subheader(design_id)

    design_interface_detail(design_id, descriptors_df, interface_descriptors_by_key)



import numpy as np
import pandas as pd
from surface_analyses.patches import assign_patches


# def get_cdrs(traj: Trajectory, cdr_def_scheme: str = 'chothia'):  # remove dependency on traj from pep-patch
def get_cdrs(sequence: str, cdr_def_scheme: str = "chothia"):
    """Adapted from Pep patch.

    Added to the hydrophobic calculation. Fixed the atom's residues to include chain id to not be ambiguous across chains.
    Returns also the number of CDR loop, not just binary CDR flag.
    Todo maybe create public PR for surface_analyses for this and multiple cutoffs?"""
    try:
        from cdr_annotation import CustomAnnotation

        return {
            # str(traj.top.residue(i)): loop_no
            res_serial: loop_no
            # for i, loop_no in CustomAnnotation.from_traj(traj, scheme=cdr_def_scheme).cdr_indices()
            for res_serial, loop_no in CustomAnnotation(sequence, scheme=cdr_def_scheme).cdr_indices()
        }
    except ImportError as e:
        print(
            f"CDR annotation failed with the following error:\n{e}\n"
            "If the error pertains to the annotation tool ANARCI or ANARCI is missing, "
            "a fresh installation of ANARCI ( https://github.com/oxpig/ANARCI ) or its dependencies might help.\n\n"
            "To use pep_patch_hydrophobic without CDR annotation, rerun the script without the '--check_cdrs' flag."
        )
        raise RuntimeError("CDR Annotation failed")


# def get_vertex_df(surf, pdb_traj: Trajectory, annotate_cdrs=True):  # remove dependency on traj from pep-patch
def get_vertex_df(surf, atom_to_residue: pd.DataFrame, sequence: str, annotate_cdrs=True):
    """Annotate vertices with CDR loop number, compute vertex area."""
    # cdrs = get_cdrs(pdb_traj) if annotate_cdrs else {}  # remove dependency on pdb_traj from pep-patch
    cdrs = get_cdrs(sequence) if annotate_cdrs else {}

    # residues = np.array([str(a.residue) for a in pdb_traj.top.atoms])
    # this is the residue serial (same as in pep-patch), first are H residues and then L residue ids start where the H left off
    residues = atom_to_residue["residue_serial"].to_numpy()

    # Calculate the area of each triangle, and split evenly among the vertices
    tri_areas = surf.areas()
    vert_areas = np.zeros(surf.vertices.shape[0])
    # Distribute the triangle areas evenly among their vertices
    np.add.at(
        vert_areas, surf.faces, tri_areas[:, np.newaxis] / 3.0
    )  # add newaxis, so tri_areas gets broadcasted to surf.faces (n,3)
    return pd.DataFrame(
        {
            "area": vert_areas,
            "atom": surf["atom"],
            "cdr": np.vectorize(cdrs.get)(residues, -1)[surf["atom"]],  # assign CDR loop no
            "value": surf["values"],
        }
    )


def assign_patches_electrostatic(vertex_potential: pd.Series, faces: np.array, patch_cutoff: tuple[float]):
    """Aggregate surface in patches using the cutoff.

    Returns a pd.DataFrame of vertices like in PEP Patch.
    Additionally, patch_id column is added, which has unique identified for each patch (unique over the union of positive and negative).
    - Vertices with *positive* charge patches have positive patch_id (1..inf)
    - Vertices with *negative* charge patches have negative patch_id (-1..-inf)
    - Vertices with no patches have patch_id set to 0

    Adapted from pep patch so we can run this with multiple patch_cutoffs while pep patch (APBS) can run just once.

    [done] And it would be great to remove the dependency with pdb_traj.., we would just need the sequence of residues for each atom and total sequence H+L, which both we have in atom_to_residue
    """
    # code from PEP-Patch
    positive_patches = assign_patches(faces, vertex_potential > patch_cutoff[0])
    negative_patches = assign_patches(faces, vertex_potential < patch_cutoff[1])

    # create column patch_id so that patch_id will be unique within positive _and_ negative patches
    patch_id = positive_patches + 1
    # renumber negative patches, so we can discern between positive and negative
    patch_id[negative_patches != -1] = -1 - negative_patches[negative_patches != -1]

    return patch_id



import argparse
import os
import glob

from biopandas.pdb import PandasPdb


def is_designed(contig_chain_or_segment: str) -> bool:
    """
    Check if the contig chain is fixed or not.
    :param contig_chain: str, the contig chain string
    :return: bool, True if designed, False if is fixed
    """
    for segment in contig_chain_or_segment.removesuffix("/0").split("/"):
        if not segment[0].isalpha():
            return True
    return False


def get_insertion_codes(pdb_path: str, contig_chains: list[str]) -> list[str]:
    """
    Get the insertion codes from the PDB files.
    :Args:
        pdb_path: str, path to the PDB file
        contig_chains: list[str], list of contig chains
    :return: A set of insertion codes.
    """

    ppdb = PandasPdb().read_pdb(pdb_path)
    chains = ppdb.df["ATOM"]["chain_id"].unique().tolist()

    residues_in_fixed_segments_by_chain = {}
    for contig_chain in contig_chains:
        for segment in contig_chain.split("/"):
            # Check if segment is designed
            if is_designed(segment):
                continue

            chain = segment[0]

            assert chain in chains, f"Chain {chain} not found in PDB file {pdb_path}. Chains found: {chains}"
            start, end = map(int, segment[1:].split("-"))

            residues_in_fixed_segments_by_chain.setdefault(chain, [])
            residues_in_fixed_segments_by_chain[chain] += list(range(start, end + 1))

    insertion_codes_residues = []
    for index, row in ppdb.df["ATOM"][["chain_id", "residue_number", "insertion"]].drop_duplicates().iterrows():
        if row["insertion"] and row["residue_number"]:
            if row["chain_id"] not in residues_in_fixed_segments_by_chain:
                continue

            if row["residue_number"] in residues_in_fixed_segments_by_chain[row["chain_id"]]:
                insertion_codes_residues.append(str(row["residue_number"]) + row["insertion"])
    return insertion_codes_residues


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("input_pdb_path", type=str)
    parser.add_argument("contig", type=str)
    options = parser.parse_args()

    if os.path.isdir(options.input_pdb_path):
        pdb_paths = sorted(glob.glob(os.path.join(options.input_pdb_path, "*.pdb")))
    else:
        pdb_paths = [options.input_pdb_path]

    print("Checking if there is a designable chain in contig")
    contig_chains = options.contig.split(" ")
    num_designed_segments = 0
    for contig_chain in contig_chains:
        num_designed_segments += sum(is_designed(segment) for segment in contig_chain.removesuffix("/0").split("/"))
    assert num_designed_segments, f"Invalid contig {options.contig}. No designed segments specified."

    print("Checking for insertion codes")
    insertion_codes_by_pdb_path = {pdb_path: get_insertion_codes(pdb_path, contig_chains) for pdb_path in pdb_paths}

    for pdb_path, insertion_codes in insertion_codes_by_pdb_path.items():
        assert not insertion_codes, (
            f"Found insertion codes {insertion_codes} in {pdb_path}. Please remove them before running the pipeline. "
        )



import time
import streamlit as st

from ovo.core.scheduler.base_scheduler import Scheduler


def wait_with_statusbar(
    scheduler: Scheduler, job_id: str, label="Workflow in progress...", interval_seconds=1.0, timeout_seconds=60 * 60
) -> bool:
    if scheduler.get_result(job_id):
        # Do not render the component at all if the job has already finished
        return True

    log_element = None
    watching_since = time.time()
    # TODO how to monitor job status in a more granular way?
    with st.status("") as statusbar:
        log_container = st.container(height=300, key=f"job_log_{job_id}")
        if st.button("Abort", key=f"abort_{job_id}"):
            statusbar.update(state="error", label="Job cancelled", expanded=False)
            scheduler.cancel(job_id)
            return False
        # Wait in sleep loop
        while True:
            result = scheduler.get_result(job_id)
            if log_text := scheduler.get_log(job_id):
                if result is None:
                    log_text += "\n...monitoring logs..."
                if log_element is None:
                    with log_container:
                        log_element = st.code(log_text)
                else:
                    log_element.code(log_text)

            if result is not None:
                if result:
                    statusbar.update(state="complete", label="Job finished")
                else:
                    statusbar.update(state="error", label="Job failed", expanded=True)
                return result

            if hasattr(scheduler, "queue_size"):
                queue_size = scheduler.queue_size(job_id=job_id)
                if queue_size is None:
                    # Job is running or queue is not used
                    statusbar.update(state="running", label=label)
                else:
                    if queue_size == 0:
                        # Job is next to run
                        behind_label = ", starting soon"
                    else:
                        # Job is waiting in front of others
                        behind_label = f" behind {queue_size} {'job' if queue_size == 1 else 'jobs'}"
                    statusbar.update(state="running", label=f"Waiting in queue{behind_label}...")

            if time.time() - watching_since > timeout_seconds:
                scheduler.cancel(job_id)
                raise TimeoutError(f"Waiting timed out after {timeout_seconds} seconds, cancelling job {job_id}")

            time.sleep(interval_seconds)



from datetime import datetime
import streamlit as st
from streamlit_timeago import time_ago

from ovo import db
from ovo.app.components.navigation import project_round_selector
from ovo.app.pages.jobs.design_job_detail import design_job_detail
from ovo.app.utils.page_init import initialize_page
from ovo.core.database.models import Pool, Round
from ovo.app.utils.cached_db import get_cached_design_jobs_table


@st.fragment
def overview():
    st.title("⏳️ Jobs")
    project = st.session_state.project

    st.write(f"Showing all jobs in project **{project.name}**")

    rounds = db.select(Round, project_id=project.id, order_by="-created_date_utc")
    if not rounds:
        st.write("*No design jobs yet in this project. Submit a workflow in the left panel.*")
        return

    selected_round_ids = project_round_selector(rounds)
    key = "_".join(sorted(selected_round_ids))

    st.subheader("Design jobs")

    st.button(":material/refresh: Refresh")
    time_ago(datetime.now(), prefix="Refreshed", key="refreshed")

    with st.spinner("Checking workflow status..."):
        table = get_cached_design_jobs_table(round_ids=selected_round_ids)
        table = table.reset_index()

    if table.empty:
        st.write("No jobs submitted yet in this round")
        return

    table.insert(
        0, ("Pool", "actions"), [f"?pool_ids={pool_id}&project_id={project.id}" for pool_id in table[("Pool", "id")]]
    )

    selection = st.dataframe(
        table,
        hide_index=True,
        key=f"pools_round_{key}",
        column_config={
            "actions": st.column_config.LinkColumn(display_text=r"Results ↑"),
        },
        on_select="rerun",
    )

    selected_ids = table[("Pool", "id")].iloc[selection["selection"]["rows"]].tolist()
    label = "All job" if not selected_ids else "Selected job"
    if st.button(f"{label} results :material/arrow_forward_ios:", key=f"open_round_{key}"):
        st.query_params["pool_ids"] = ",".join(selected_ids if selected_ids else table[("Pool", "id")])
        st.rerun()


if "pool_ids" in st.query_params:
    assert "project_id" in st.query_params, "Invalid URL, project ID expected"
    # Get the pool(s)
    pool_ids = st.query_params.get("pool_ids").split(",")
    project_id = st.query_params.get("project_id")
    pools = db.select(Pool, id__in=pool_ids)
    if not pools:
        raise ValueError("Pool not found")

    # check the project ID for security reasons, to avoid users guessing the short pool ID
    for round_id in set([p.round_id for p in pools]):
        project_round = db.get(Round, round_id)
        assert project_round.project_id == project_id, f"Pool does not belong to project {project_id}"

    name = f"{len(pools)} pools" if len(pools) > 1 else pools[0].name
    initialize_page(f"{name} - Jobs")
    if st.button(":material/arrow_back_ios: Back to jobs"):
        del st.query_params["pool_ids"]
        st.rerun()
    design_job_detail([pool.id for pool in pools])

# Overview page
else:
    initialize_page("Jobs")
    overview()



import glob
import os
from typing import Tuple, Dict
from Bio import SeqIO
from Bio.SeqUtils.ProtParam import ProteinAnalysis
from Bio.PDB import PDBParser
from Bio.SeqUtils import seq1
import pandas as pd
import numpy as np
import argparse
from scipy.stats import entropy

AMINO_ACIDS = list("ACDEFGHIKLMNPQRSTVWY")
NONPOLAR = list("GAVLI")
POLAR_UNCHARGED = list("STCNQY")
POSITIVE = list("KRH")
NEGATIVE = list("DE")
TURN = list("NGPS")  # (Asn, Gly, Pro and Ser)
AROMATIC = list("FYW")
DENQ = list("DENQ")
GROUPS = [NONPOLAR, POLAR_UNCHARGED, POSITIVE, NEGATIVE, TURN, AROMATIC, DENQ]

pdb_parser: PDBParser = PDBParser(QUIET=True)


def get_sequence_from_pdb_file(file_path: str, chains: list[str]) -> str:
    """
    Get the sequence of a structure from the pdb file.
    """
    with open(file_path) as f:
        parsed_structure = pdb_parser.get_structure("", f)
        sequence = []
        for chain in chains:
            for residue in parsed_structure[0][chain]:
                if residue.has_id("CA"):
                    sequence.append(seq1(residue.get_resname()))
        return "".join(sequence)


def get_protein_analysis_df(sequences_by_ids: Dict[str, str]) -> pd.DataFrame:
    """
    Get a DataFrame of sequence composition (relative abundance of each aa and aa group) and the output file of ProteinAnalysis.
    """
    df = pd.DataFrame({"sequence": sequences_by_ids})
    df.index.name = "id"

    AMINO_ACIDS_perc = [f"{aa}_perc" for aa in AMINO_ACIDS]
    df["length"] = df["sequence"].apply(len)
    df[AMINO_ACIDS_perc] = (
        df.apply(lambda x: pd.Series([x["sequence"].count(aa) / x["length"] * 100 for aa in AMINO_ACIDS]), axis=1)
        if not df.empty
        else None
    )
    cols = [
        "nonpolar_perc",
        "polar_uncharged_perc",
        "positive_perc",
        "negative_perc",
        "turn_forming_perc",
        "aromatic_perc",
        "denq_perc",
    ]
    df[cols] = (
        df.apply(
            lambda x: pd.Series(
                [np.sum([x["sequence"].count(a) for a in group]) / x["length"] * 100 for group in GROUPS]
            ),
            axis=1,
        )
        if not df.empty
        else None
    )

    protein_analysis = df["sequence"].apply(lambda x: ProteinAnalysis(x))
    df["aromaticity"] = protein_analysis.apply(lambda x: safe_aromaticity(x))
    df["charge_5_5"] = protein_analysis.apply(lambda x: safe_charge(x, 5.5))
    df["charge_7_4"] = protein_analysis.apply(lambda x: safe_charge(x, 7.4))
    df["isoelectric_point"] = protein_analysis.apply(lambda x: safe_ip(x))
    df[["MEC_reduced", "MEC_cystines"]] = (
        protein_analysis.apply(lambda x: pd.Series(safe_mec(x))) if not df.empty else None
    )
    df[["helix_perc", "turn_perc", "sheet_perc"]] = (
        protein_analysis.apply(lambda x: pd.Series(safe_sec_str(x))) if not df.empty else None
    )

    sequence_wo_X = df["sequence"].apply(lambda x: "".join([aa for aa in list(x) if aa != "X"]))
    protein_analysis_wo_X = sequence_wo_X.apply(lambda x: ProteinAnalysis(x))
    df["flexibility_avg"] = protein_analysis_wo_X.apply(lambda x: safe_flexibility_avg(x))
    df["gravy"] = protein_analysis_wo_X.apply(lambda x: safe_gravy(x))
    df["instability_index"] = protein_analysis_wo_X.apply(lambda x: safe_instability_index(x))
    df["molecular_weight"] = protein_analysis_wo_X.apply(lambda x: safe_molecular_weight(x))
    df["avg_entropy"] = df["sequence"].apply(lambda x: average_entropy(x))

    return df


def get_charge_perc(x) -> Tuple[float, float]:
    """
    Calculate the percentage of positive and negative charges in a list of charges.
    """
    x_wo_none = [value for value in x if value is not None]
    positive_perc = len([value for value in x_wo_none if value > 0]) / len(x_wo_none)
    negative_perc = len([value for value in x_wo_none if value < 0]) / len(x_wo_none)
    return (positive_perc, negative_perc)


def charged_avg(x) -> Tuple[float, float]:
    """
    Calculate the average positive and negative charges in a list of charges.
    """
    x_wo_none = [value for value in x if value is not None]
    positive_avg = np.mean([value for value in x_wo_none if value > 0])
    negative_avg = np.mean([value for value in x_wo_none if value < 0])
    return (positive_avg, negative_avg)


def safe_aromaticity(analysis: ProteinAnalysis) -> float:
    """
    Calculate the aromaticity - relative frequency of Phe+Trp+Tyr.
    """
    try:
        return analysis.aromaticity()
    except Exception:
        return np.nan


def safe_charge(analysis: ProteinAnalysis, pH=5.5) -> float:
    """
    Calculate the charge of a protein at given pH.
    """
    try:
        return analysis.charge_at_pH(pH)
    except Exception:
        return np.nan


def safe_ip(analysis: ProteinAnalysis) -> float:
    """
    Calculate the isoelectric point.
    """
    try:
        return analysis.isoelectric_point()
    except Exception:
        return np.nan


def safe_mec(analysis: ProteinAnalysis) -> Tuple[int, int]:
    """
    Molar extinction coefficient assuming cysteines (reduced) and cystines residues (Cys-Cys-bond).
    """
    try:
        return analysis.molar_extinction_coefficient()
    except Exception:
        return (np.nan, np.nan)


def safe_sec_str(analysis: ProteinAnalysis) -> Tuple[float, float, float]:
    """
    Calculate percentage of helix, turn and sheet.
    """
    try:
        return [frac * 100 for frac in analysis.secondary_structure_fraction()]
    except Exception:
        return (np.nan, np.nan, np.nan)


def safe_flexibility_avg(analysis: ProteinAnalysis) -> float:
    """
    Calculates the average flexibility of the protein.
    """
    try:
        return np.mean(analysis.flexibility())
    except Exception:
        return np.nan


def safe_gravy(analysis: ProteinAnalysis) -> float:
    """
    Calculate the GRAVY (Grand Average of Hydropathy) value.
    """
    try:
        return analysis.gravy()
    except Exception:
        return np.nan


def safe_instability_index(analysis: ProteinAnalysis) -> float:
    """
    Calculate the instability index.
    """
    try:
        return analysis.instability_index()
    except Exception:
        return np.nan


def safe_molecular_weight(analysis: ProteinAnalysis) -> float:
    """
    Calculate the molecular weight from protein sequence.
    """
    try:
        return analysis.molecular_weight()
    except Exception:
        return np.nan


def aa_entropy(sequence: str) -> float:
    """
    Calculate the amino acid entropy of a sequence.
    Non-standard amino acids (B, J, O, U, X, Z) are ignored.
    """
    counts = pd.Series(list(sequence)).value_counts().to_dict()
    return entropy([counts.get(aa, 0) for aa in AMINO_ACIDS], base=2)


def windowed_aa_entropy(sequence: str, window_size: int) -> float:
    """
    Calculate the average amino acid entropy in sliding windows of the sequence.
    """
    entropy = (
        pd.Series([sequence[i : i + window_size] for i in range(len(sequence) - (window_size - 1))])
        .apply(aa_entropy)
        .mean()
    )
    return entropy


def average_entropy(sequence, window_size=21):
    """
    Calculate the amino acid entropy of a sequence, non-standard amino acids (B, J, O, U, X, Z) are ignored.
    1. If the sequence is shorter than the window size, return the entropy of the whole sequence.
    2. Otherwise, return the average entropy of sliding windows of the given size (= 21 by default to match ProteinSol implementation).
    """
    if len(sequence) < window_size:
        return aa_entropy(sequence)
    return windowed_aa_entropy(sequence, window_size)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("input_path", type=str)
    parser.add_argument("output_csv", type=str)
    parser.add_argument(
        "-c", "--chains", type=str, required=True, help="Chain(s) to extract the sequence from, comma-separated"
    )
    options = parser.parse_args()

    chains = list(options.chains.replace(" ", "").replace(",", ""))
    print("Chains:", chains)

    if os.path.isdir(options.input_path):
        paths = sorted(glob.glob(os.path.join(options.input_path, "*.pdb")))
        print(f"Reading sequences from {len(paths):,} PDBs")
        sequences_by_id = {
            os.path.splitext(os.path.basename(path))[0]: get_sequence_from_pdb_file(path, chains=chains)
            for path in paths
        }
    elif options.input_path.endswith((".pdb",)):
        sequences_by_id = {
            os.path.splitext(os.path.basename(options.input_path))[0]: get_sequence_from_pdb_file(
                options.input_path, chains=chains
            )
        }
    elif options.input_path.endswith((".fasta", ".fa")):
        sequences_by_id = {record.id: str(record.seq) for record in SeqIO.parse(options.input_path, "fasta")}
    else:
        raise ValueError("Input must be a directory with PDB files, a PDB file, or a FASTA file")
    print(f"Calculating sequence composition on {len(sequences_by_id):,} sequences")
    df = get_protein_analysis_df(sequences_by_id)
    print(df)
    df.to_csv(options.output_csv)
    print("Saved to:", options.output_csv)



#!/usr/bin/env python

import os
import sys

from pyrosetta import *
from pyrosetta.rosetta import *

import numpy as np
from collections import OrderedDict
import time
import argparse
import subprocess
import glob
import random
import torch
import json

import util_protein_mpnn as mpnn_util

parent = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(os.path.join(parent, "include"))
from silent_tools import silent_tools


def cmd(command, wait=True):
    the_command = subprocess.Popen(
        command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True
    )
    if not wait:
        return
    the_stuff = the_command.communicate()
    return str(the_stuff[0]) + str(the_stuff[1])


def range1(iterable):
    return range(1, iterable + 1)


#################################
# Parse Arguments
#################################

parser = argparse.ArgumentParser()
script_dir = os.path.dirname(os.path.realpath(__file__))

# I/O Arguments
parser.add_argument("-pdbdir", type=str, default="", help="The name of a directory of pdbs to run through the model")
parser.add_argument("-silent", type=str, default="", help="The name of a silent file to run through the model")
parser.add_argument(
    "-outpdbdir",
    type=str,
    default="outputs",
    help="The directory to which the output PDB files will be written, used if the -pdbdir arg is active",
)
parser.add_argument(
    "-outsilent",
    type=str,
    default="out.silent",
    help="The name of the silent file to which output structs will be written, used if the -silent arg is active",
)
parser.add_argument(
    "-runlist",
    type=str,
    default="",
    help="The path of a list of pdb tags to run, only active when the -pdbdir arg is active (default: ''; Run all PDBs)",
)
parser.add_argument(
    "-checkpoint_name",
    type=str,
    default="check.point",
    help="The name of a file where tags which have finished will be written (default: check.point)",
)

parser.add_argument(
    "-debug",
    action="store_true",
    default=False,
    help="When active, errors will cause the script to crash and the error message to be printed out (default: False)",
)

# Design Arguments
parser.add_argument(
    "-relax_cycles", type=int, default=1, help="The number of relax cycles to perform on each structure (default: 1)"
)
parser.add_argument(
    "-output_intermediates",
    action="store_true",
    help="Whether to write all intermediate sequences from the relax cycles to disk (default: False)",
)
parser.add_argument(
    "-seqs_per_struct",
    type=int,
    default="1",
    help="The number of sequences to generate for each structure (default: 1)",
)

# ProteinMPNN-Specific Arguments
parser.add_argument(
    "-checkpoint_path",
    type=str,
    default=os.path.join(script_dir, "ProteinMPNN/vanilla_model_weights/v_48_020.pt"),
    help=f"The path to the ProteinMPNN weights you wish to use, default {os.path.join(script_dir, 'ProteinMPNN/vanilla_model_weights/v_48_020.pt')}",
)
parser.add_argument(
    "-temperature",
    type=float,
    default=0.000001,
    help="The sampling temperature to use when running ProteinMPNN (default: 0.000001)",
)
parser.add_argument(
    "-augment_eps",
    type=float,
    default=0,
    help="The variance of random noise to add to the atomic coordinates (default 0)",
)
parser.add_argument(
    "-protein_features",
    type=str,
    default="full",
    help="What type of protein features to input to ProteinMPNN (default: full)",
)
parser.add_argument(
    "-omit_AAs",
    type=str,
    default="CX",
    help="A string of all residue types (one letter case-insensitive) that you would not like to use for design. Letters not corresponding to residue types will be ignored (default: CX)",
)
parser.add_argument(
    "-bias_AA_jsonl",
    type=str,
    default="",
    help="The path to a JSON file containing a dictionary mapping residue one-letter names to the bias for that residue eg. {A: -1.1, F: 0.7} (default: "
    "; no bias)",
)
parser.add_argument(
    "-bias_AA",
    type=str,
    default="",
    help="String representation mapping residue one-letter names to the bias for that residue eg. A:-1.1,F:0.7 (default: no bias)",
)
parser.add_argument(
    "-num_connections",
    type=int,
    default=48,
    help="Number of neighbors each residue is connected to. Do not mess around with this argument unless you have a specific set of ProteinMPNN weights which expects a different number of connections. (default: 48)",
)
parser.add_argument("-seed", type=int, default=None, help="Random seed - DANGER - will produce same output each time")

args = parser.parse_args(sys.argv[1:])

init_args = "-beta_nov16 -in:file:silent_struct_type binary -mute basic.io.database core.scoring"

if not args.debug:
    init_args += " -mute all"

xml_path = os.path.join(script_dir, "RosettaFastRelaxUtil.xml")
init_args += " -use_terminal_residues true"

if args.seed:
    # Deterministic mode using a fixed random seed
    torch.manual_seed(args.seed)
    random.seed(args.seed)
    np.random.seed(args.seed)
    init_args += f" -constant_seed -jran {args.seed}"

# Initialize PyRosetta
init(init_args)


class sample_features:
    """
    This is a struct which keeps all the features related to a single sample together
    """

    def __init__(self, pose, tag):
        self.pose = pose
        self.tag = os.path.basename(tag).split(".")[0]

    def parse_fixed_res(self):
        """
        Parse which residues are fixed from the residue remarks
        """

        # Iterate over the residue positions in this pose and check if they are fixed
        fixed_list = []

        endA = self.pose.split_by_chain()[1].total_residue()
        for resi in range1(endA):
            reslabels = self.pose.pdb_info().get_reslabels(resi)

            if len(reslabels) == 0:
                continue

            if str(self.pose.pdb_info().get_reslabels(resi)[1]).strip() == "FIXED":
                fixed_list.append(resi)
        # Get ordered unique chainIDs for the pose
        # The last chain (B) is the target and will have a fixed sequence
        self.chains = list(
            OrderedDict.fromkeys([self.pose.pdb_info().chain(i) for i in range1(self.pose.total_residue())])
        )

        # Create the fixed res dict, this will be input to ProteinMPNN
        self.fixed_res = {self.chains[0]: fixed_list, self.chains[1]: []}

    def thread_mpnn_seq(self, binder_seq):
        """
        Thread the binder sequence onto the pose being designed
        """
        rsd_set = self.pose.residue_type_set_for_pose(core.chemical.FULL_ATOM_t)

        for resi, mut_to in enumerate(binder_seq):
            resi += 1  # 1 indexing
            name3 = mpnn_util.aa_1_3[mut_to]
            new_res = core.conformation.ResidueFactory.create_residue(rsd_set.name_map(name3))
            self.pose.replace_residue(resi, new_res, True)


class ProteinMPNN_runner:
    """
    This class is designed to run the ProteinMPNN model on a single input. This class handles
    the loading of the model, the loading of the input data, the FastRelax cycles, and the processing of the output
    """

    def __init__(self, args, struct_manager):
        self.struct_manager = struct_manager

        if torch.cuda.is_available():
            print("Found GPU will run ProteinMPNN on GPU")
            self.device = "cuda:0"
        else:
            print("No GPU found, running ProteinMPNN on CPU")
            self.device = "cpu"

        self.mpnn_model = mpnn_util.init_seq_optimize_model(
            self.device,
            hidden_dim=128,
            num_layers=3,
            backbone_noise=args.augment_eps,
            num_connections=args.num_connections,
            checkpoint_path=args.checkpoint_path,
        )

        alphabet = "ACDEFGHIKLMNPQRSTVWYX"

        self.debug = args.debug

        self.temperature = args.temperature
        self.seqs_per_struct = args.seqs_per_struct
        self.omit_AAs = [letter for letter in args.omit_AAs.upper() if letter in list(alphabet)]

        # Parse AA bias settings from json
        if args.bias_AA_jsonl or args.bias_AA:
            if args.bias_AA_jsonl:
                assert not args.bias_AA, "Cannot use both -bias_AA_jsonl and -bias_AA at the same time"
                print(f"Found AA bias json file at {args.bias_AA_jsonl}")
                with open(args.bias_AA_jsonl, "r") as json_file:
                    json_list = list(json_file)
                for json_str in json_list:
                    bias_AA_dict = json.loads(json_str)
            else:
                print(f"Parsing AA bias from string {args.bias_AA}")
                bias_AA_dict = {}
                for entry in args.bias_AA.split(","):
                    if len(entry.split(":")) != 2:
                        raise Exception(f"Expected bias format: A:-1.1,F:0.7 but got {entry}")
                    AA, bias = entry.split(":")
                    if AA not in list(alphabet):
                        raise Exception(f"AA {AA} in -bias_AA argument is not a valid amino acid")
                    try:
                        bias = float(bias)
                    except:
                        raise Exception(f"Cannot convert bias {bias} to float in -bias_AA argument")
                    bias_AA_dict[AA] = bias

            self.bias_AAs_np = np.zeros(len(alphabet))
            for n, AA in enumerate(alphabet):
                if AA in list(bias_AA_dict.keys()):
                    self.bias_AAs_np[n] = bias_AA_dict[AA]
        else:
            # Not using AA bias
            self.bias_AAs_np = np.zeros(len(alphabet))

        # Configs for the FastRelax cycles
        self.relax_cycles = args.relax_cycles

    def relax_pose(self, mover, sample_feats):
        """
        Run FastRelax on the current pose
        """

        relaxT0 = time.time()

        print("Running FastRelax")

        mover.apply(sample_feats.pose)

        print(f"Completed one cycle of FastRelax in {int(time.time()) - relaxT0} seconds")

    def sequence_optimize(self, sample_feats):
        mpnn_t0 = time.time()

        # Once we have figured out pose I/O without Rosetta this will be easy to swap in
        pdbfile = "temp.pdb"
        sample_feats.pose.dump_pdb(pdbfile)

        feature_dict = mpnn_util.generate_seqopt_features(pdbfile, sample_feats.chains)

        os.remove(pdbfile)

        arg_dict = mpnn_util.set_default_args(self.seqs_per_struct, omit_AAs=self.omit_AAs)
        arg_dict["temperature"] = self.temperature

        masked_chains = sample_feats.chains[:-1]
        visible_chains = [sample_feats.chains[-1]]

        fixed_positions_dict = {pdbfile[: -len(".pdb")]: sample_feats.fixed_res}

        if self.debug:
            print(f"Fixed positions dict: {fixed_positions_dict}")

        sequences = mpnn_util.generate_sequences(
            self.mpnn_model,
            self.device,
            feature_dict,
            arg_dict,
            masked_chains,
            visible_chains,
            bias_AAs_np=self.bias_AAs_np,
            fixed_positions_dict=fixed_positions_dict,
        )

        if self.debug:
            print(f"Generated sequence(s): {sequences}")

        print(f"ProteinMPNN generated {len(sequences)} sequences in {int(time.time() - mpnn_t0)} seconds")

        return sequences

    def proteinmpnn(self, sample_feats):
        """
        Run ProteinMPNN sequence optimization on the pose, this does not use FastRelax
        """
        seqs_scores = self.sequence_optimize(sample_feats)

        # Iterate though each seq score pair and thread the sequence onto the pose
        # Then write each pose to a pdb file
        prefix = f"{sample_feats.tag}_dldesign"
        for idx, (seq, score) in enumerate(seqs_scores):
            sample_feats.thread_mpnn_seq(seq)

            outtag = f"{prefix}_{idx}"

            self.struct_manager.dump_pose(sample_feats.pose, outtag)

    def proteinmpnn_fastrelax(self, sample_feats):
        """
        Run ProteinMPNN plus FastRelax on the pose being designed
        """
        tot_t0 = time.time()
        design_counter = 0

        # We re-initialize the mover for each PDB to avoid remembering stuff from previous peptides, like peptide length
        objs = protocols.rosetta_scripts.XmlObjects.create_from_file(xml_path)
        # ParsedProtocol corresponds to PROTOCOLS block in the XML
        mover = objs.get_mover("ParsedProtocol")

        prefix = f"{sample_feats.tag}_dldesign"

        for cycle in range(args.relax_cycles):
            seqs_scores = self.sequence_optimize(sample_feats)

            seq, mpnn_score = seqs_scores[0]  # We know there is only one entry
            sample_feats.thread_mpnn_seq(seq)

            self.relax_pose(mover, sample_feats)

            if args.output_intermediates:
                tag = f"{prefix}_0_cycle{cycle}"
                self.struct_manager.dump_pose(sample_feats.pose, tag)

        # NOTE this is a change compared to the dl_binder_design repository
        # We do not perform this last sequence design - this ensures all designs are relaxed,
        # which enables running ddG calculation directly on the designed structure
        # Designs from intermediate cycles were already relaxed so this ensures consistency.
        #
        # seqs_scores = self.sequence_optimize(sample_feats)
        #
        # seq, mpnn_score = seqs_scores[0]  # We know there is only one entry
        # sample_feats.thread_mpnn_seq(seq)
        #
        # tag = f"{prefix}_0_cycle{args.relax_cycles}"
        # self.struct_manager.dump_pose(sample_feats.pose, tag)

    def run_model(self, tag, args):
        t0 = time.time()

        # Load the pose
        pose = self.struct_manager.load_pose(tag)

        # Initialize the features
        sample_feats = sample_features(pose, tag)

        # Parse the fixed residues from the pose remarks
        sample_feats.parse_fixed_res()

        # Now determine which type of run we are doing and execute it
        if args.relax_cycles > 0:
            if args.seqs_per_struct > 1:
                raise Exception("Cannot use -seqs_per_struct > 1 with -relax_cycles > 0")

            self.proteinmpnn_fastrelax(sample_feats)

        else:
            self.proteinmpnn(sample_feats)

        seconds = int(time.time() - t0)

        print(f"Struct: {pdb} reported success in {seconds} seconds")


class StructManager:
    """
    This class handles all of the input and output for the ProteinMPNN model. It deals with silent files vs. pdbs,
    checkpointing, and writing of outputs

    Note: This class could be moved to a separate file
    """

    def __init__(self, args):
        self.args = args

        self.silent = False
        if not args.silent == "":
            self.silent = True

            self.struct_iterator = silent_tools.get_silent_index(args.silent)["tags"]

            self.sfd_in = rosetta.core.io.silent.SilentFileData(rosetta.core.io.silent.SilentFileOptions())
            self.sfd_in.read_file(args.silent)

            self.sfd_out = core.io.silent.SilentFileData(
                args.outsilent, False, False, "binary", core.io.silent.SilentFileOptions()
            )

            self.outsilent = args.outsilent

        self.pdb = False
        if not args.pdbdir == "":
            self.pdb = True

            self.pdbdir = args.pdbdir
            self.outpdbdir = args.outpdbdir

            self.struct_iterator = sorted(glob.glob(os.path.join(args.pdbdir, "*.pdb")))

            if not os.path.exists(self.outpdbdir):
                os.makedirs(self.outpdbdir)

            # Parse the runlist and determine which structures to process
            if args.runlist != "":
                with open(args.runlist, "r") as f:
                    self.runlist = set([line.strip() for line in f])

                    # Filter the struct iterator to only include those in the runlist
                    self.struct_iterator = [
                        struct
                        for struct in self.struct_iterator
                        if os.path.basename(struct).split(".")[0] in self.runlist
                    ]

                    print(f"After filtering by runlist, {len(self.struct_iterator)} structures remain")

        # Assert that either silent or pdb is true, but not both
        assert self.silent ^ self.pdb, (
            f"Both silent and pdb are set to {args.silent} and {args.pdb} respectively. Only one of these may be active at a time"
        )

        # Setup checkpointing
        self.chkfn = args.checkpoint_name
        self.finished_structs = set()

        if os.path.isfile(self.chkfn):
            with open(self.chkfn, "r") as f:
                for line in f:
                    self.finished_structs.add(line.strip())

    def record_checkpoint(self, tag):
        """
        Record the fact that this tag has been processed.
        Write this tag to the list of finished structs
        """
        with open(self.chkfn, "a") as f:
            f.write(f"{tag}\n")

    def iterate(self):
        """
        Iterate over the silent file or pdb directory and run the model on each structure
        """

        # Iterate over the structs and for each, check that the struct has not already been processed
        for i, struct in enumerate(self.struct_iterator, start=1):
            tag = os.path.basename(struct).split(".")[0]
            if tag in self.finished_structs:
                print(f"{tag} has already been processed. Skipping")
                continue
            print(f"Processing PDB {i:,}/{len(self.struct_iterator):,}: {tag}")

            yield struct

    def dump_pose(self, pose, tag):
        """
        Dump this pose to either a silent file or a pdb file depending on the input arguments
        """
        if self.pdb:
            pdbfile = os.path.join(self.outpdbdir, tag + ".pdb")
            pose.dump_pdb(pdbfile)

        if self.silent:
            struct = self.sfd_out.create_SilentStructOP()
            struct.fill_struct(pose, tag)

            self.sfd_out.add_structure(struct)
            self.sfd_out.write_silent_struct(struct, self.outsilent)

    def load_pose(self, tag):
        """
        Load a pose from either a silent file or a pdb file depending on the input arguments
        """

        if not self.pdb and not self.silent:
            raise Exception("Neither pdb nor silent is set to True. Cannot load pose")

        if self.pdb:
            pose = pose_from_pdb(tag)

        if self.silent:
            pose = Pose()
            self.sfd_in.get_structure(tag).fill_pose(pose)

        return pose


####################
####### Main #######
####################

struct_manager = StructManager(args)
proteinmpnn_runner = ProteinMPNN_runner(args, struct_manager)

for pdb in struct_manager.iterate():
    if args.debug:
        proteinmpnn_runner.run_model(pdb, args)

    else:  # When not in debug mode the script will continue to run even when some poses fail
        t0 = time.time()

        try:
            proteinmpnn_runner.run_model(pdb, args)

        except KeyboardInterrupt:
            sys.exit("Script killed by Control+C, exiting")

        except:
            seconds = int(time.time() - t0)
            print("Struct with tag %s failed in %i seconds with error: %s" % (pdb, seconds, sys.exc_info()[0]))

    # We are done with one pdb, record that we finished
    struct_manager.record_checkpoint(pdb)



from ovo.core.database.models import NumericGlobalDescriptor, ResidueNumberDescriptor

DESCRIPTORS = [
    NumericGlobalDescriptor(
        name="MPNN score",
        description="MPNN sequence score, generally not recommended for filtering as it depends on protein",
        tool="BindCraft",
        key="bindcraft|mpnn|MPNN_score",
        comparison="does_not_apply",
    ),
    NumericGlobalDescriptor(
        name="MPNN seq recovery",
        description="MPNN sequence recovery of original trajectory",
        tool="BindCraft",
        key="bindcraft|mpnn|MPNN_seq_recovery",
        comparison="does_not_apply",
    ),
    NumericGlobalDescriptor(
        name="AF2 pLDDT",
        description="pLDDT confidence score of AF2 complex prediction (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|af2|Average_pLDDT",
        comparison="higher_is_better",
        min_value=0,
        color_scale="plddt",
    ),
    NumericGlobalDescriptor(
        name="AF2 pTM score",
        description="pTM confidence score of AF2 complex prediction, 0 = worst, 1 = best (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|af2|Average_pTM",
        comparison="higher_is_better",
        min_value=0,
    ),
    NumericGlobalDescriptor(
        name="AF2 Interface pTM score",
        description="Interface pTM confidence score of AF2 complex prediction, 0 = worst, 1 = best (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|af2|Average_i_pTM",
        comparison="higher_is_better",
        min_value=0,
    ),
    NumericGlobalDescriptor(
        name="AF2 pAE",
        description="Predicted alignment error of AF2 complex prediction, in Angstrom (average across AF2 models)",
        unit="Å",
        tool="BindCraft",
        key="bindcraft|af2|Average_pAE",
        comparison="lower_is_better",
        min_value=0,
    ),
    NumericGlobalDescriptor(
        name="AF2 Interface PAE",
        description="Predicted interface alignment error of AF2 complex prediction, in Angstrom (average across AF2 models)",
        unit="Å",
        tool="BindCraft",
        key="bindcraft|af2|Average_i_pAE",
        comparison="lower_is_better",
        min_value=0,
    ),
    NumericGlobalDescriptor(
        name="AF2 Interface pLDDT",
        description="Interface pLDDT confidence score of AF2 complex prediction (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|af2|Average_i_pLDDT",
        comparison="higher_is_better",
        min_value=0,
        color_scale="plddt",
    ),
    NumericGlobalDescriptor(
        name="AF2 Secondary Structure pLDDT",
        description="Secondary structure pLDDT confidence score of AF2 complex prediction (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|af2|Average_ss_pLDDT",
        comparison="higher_is_better",
        min_value=0,
        color_scale="plddt",
    ),
    NumericGlobalDescriptor(
        name="Unrelaxed Clashes",
        description="Number of interface clashes before relaxation (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|interface|Average_Unrelaxed_Clashes",
        comparison="does_not_apply",
    ),
    NumericGlobalDescriptor(
        name="Relaxed Clashes",
        description="Number of interface clashes after relaxation (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|interface|Average_Relaxed_Clashes",
        comparison="does_not_apply",
    ),
    NumericGlobalDescriptor(
        name="PyRosetta Binder Energy Score",
        description="Rosetta energy score for binder alone (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|interface|Average_Binder_Energy_Score",
        comparison="does_not_apply",
    ),
    NumericGlobalDescriptor(
        name="Surface Hydrophobicity",
        description="Fraction of hydrophobic residues out of all surface-exposed residues (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|interface|Average_Surface_Hydrophobicity",
        comparison="does_not_apply",
    ),
    NumericGlobalDescriptor(
        name="Shape Complementarity",
        description="Interface shape complementarity (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|interface|Average_ShapeComplementarity",
        comparison="does_not_apply",
    ),
    NumericGlobalDescriptor(
        name="PackStat",
        description="Interface packstat rosetta score (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|interface|Average_PackStat",
        comparison="does_not_apply",
    ),
    NumericGlobalDescriptor(
        name="PyRosetta Interface dG",
        description="Interface rosetta dG (complex total residue energy - separated total energy) calculated with InterfaceAnalyzerMover default full-atom energy terms (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|interface|Average_dG",
        comparison="does_not_apply",
    ),
    NumericGlobalDescriptor(
        name="Delta SASA",
        description="Interface delta SASA (size) (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|interface|Average_dSASA",
        comparison="does_not_apply",
    ),
    NumericGlobalDescriptor(
        name="PyRosetta dG/dSASA",
        description="Interface energy (complex total residue energy - separated total energy) divided by interface size (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|interface|Average_dG/dSASA",
        comparison="does_not_apply",
    ),
    NumericGlobalDescriptor(
        name="Interface SASA %",
        description="Fraction of binder surface covered by the interface (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|interface|Average_Interface_SASA_%",
        comparison="does_not_apply",
    ),
    NumericGlobalDescriptor(
        name="Interface Hydrophobicity",
        description="Interface hydrophobicity fraction of binder interface (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|interface|Average_Interface_Hydrophobicity",
        comparison="does_not_apply",
    ),
    NumericGlobalDescriptor(
        name="Num Interface Residues",
        description="Number of interface residues (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|interface|Average_n_InterfaceResidues",
        comparison="does_not_apply",
    ),
    NumericGlobalDescriptor(
        name="Num Interface Hbonds",
        description="Number of hydrogen bonds at the interface (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|interface|Average_n_InterfaceHbonds",
        comparison="does_not_apply",
    ),
    NumericGlobalDescriptor(
        name="Interface Hbonds %",
        description="Number of hydrogen bonds compared to interface size (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|interface|Average_InterfaceHbondsPercentage",
        comparison="does_not_apply",
    ),
    NumericGlobalDescriptor(
        name="N Interface Unsat Hbonds",
        description="Number of unsatisfied buried hydrogen bonds at the interface (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|interface|Average_n_InterfaceUnsatHbonds",
        comparison="does_not_apply",
    ),
    NumericGlobalDescriptor(
        name="Interface Unsat Hbonds %",
        description="Number of unsatisfied buried hydrogen bonds compared to interface size (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|interface|Average_InterfaceUnsatHbondsPercentage",
        comparison="does_not_apply",
    ),
    NumericGlobalDescriptor(
        name="Interface Helix %",
        description="Proportion of alfa helices at the interface (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|dssp|Average_Interface_Helix%",
        comparison="does_not_apply",
        min_value=0,
        max_value=100,
    ),
    NumericGlobalDescriptor(
        name="Interface Beta Sheet %",
        description="Proportion of beta sheets at the interface (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|dssp|Average_Interface_BetaSheet%",
        comparison="does_not_apply",
        min_value=0,
        max_value=100,
    ),
    NumericGlobalDescriptor(
        name="Interface Loop %",
        description="Proportion of loops at the interface (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|dssp|Average_Interface_Loop%",
        comparison="does_not_apply",
        min_value=0,
        max_value=100,
    ),
    NumericGlobalDescriptor(
        name="Binder Helix %",
        description="Proportion of alfa helices in the binder structure (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|dssp|Average_Binder_Helix%",
        comparison="does_not_apply",
        min_value=0,
        max_value=100,
    ),
    NumericGlobalDescriptor(
        name="Binder BetaSheet %",
        description="Proportion of beta sheets in the binder structure (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|dssp|Average_Binder_BetaSheet%",
        comparison="does_not_apply",
        min_value=0,
        max_value=100,
    ),
    NumericGlobalDescriptor(
        name="Binder Loop %",
        description="proportion of loops in the binder structure (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|dssp|Average_Binder_Loop%",
        comparison="does_not_apply",
        min_value=0,
        max_value=100,
    ),
    NumericGlobalDescriptor(
        name="AF2 Target-aligned Binder RMSD",
        description='RMSD of binder aligned on target compared to original trajectory, in other words how far is binder in the repredicted complex from the original binding site. Called "AF2 Hotspot RMSD" in BindCraft. (average across AF2 models)',
        unit="Å",
        tool="BindCraft",
        key="bindcraft|af2|Average_Hotspot_RMSD",
        comparison="lower_is_better",
        min_value=0,
        color_scale="rmsd",
    ),
    NumericGlobalDescriptor(
        name="AF2 Target RMSD",
        description="RMSD of target predicted in context of the designed binder compared to input PDB (average across AF2 models)",
        unit="Å",
        tool="BindCraft",
        key="bindcraft|af2|Average_Target_RMSD",
        comparison="lower_is_better",
        min_value=0,
        color_scale="rmsd",
    ),
    NumericGlobalDescriptor(
        name="AF2 Binder pLDDT",
        description="pLDDT confidence score of binder predicted alone (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|af2|Average_Binder_pLDDT",
        comparison="higher_is_better",
        min_value=0,
        color_scale="plddt",
    ),
    NumericGlobalDescriptor(
        name="AF2 Binder pTM score",
        description="pTM confidence score of binder predicted alone (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|af2|Average_Binder_pTM",
        comparison="higher_is_better",
        min_value=0,
    ),
    NumericGlobalDescriptor(
        name="AF2 Binder pAE",
        description="predicted alignment error of binder predicted alone (average across AF2 models)",
        tool="BindCraft",
        key="bindcraft|af2|Average_Binder_pAE",
        comparison="lower_is_better",
        min_value=0,
    ),
    NumericGlobalDescriptor(
        name="AF2 Binder-aligned Binder RMSD",
        description="RMSD of binder predicted alone compared to original trajectory (average across AF2 models)",
        unit="Å",
        tool="BindCraft",
        key="bindcraft|af2|Average_Binder_RMSD",
        comparison="lower_is_better",
        min_value=0,
        color_scale="rmsd",
    ),
    ResidueNumberDescriptor(
        name="Binder Interface Residues",
        description="Comma-separated list of binder residue numbers on the interface",
        tool="BindCraft",
        key="bindcraft|interface|InterfaceResidues",
    ),
]

_descriptors_by_key = {desc.key: desc for desc in DESCRIPTORS}

PRESETS = [
    {
        "label": "BindCraft Interface PAE & Binder pLDDT",
        "x": _descriptors_by_key["bindcraft|af2|Average_i_pAE"],
        "y": _descriptors_by_key["bindcraft|af2|Average_Binder_pLDDT"],
    },
    {
        "label": "BindCraft Helix vs Sheet",
        "x": _descriptors_by_key["bindcraft|dssp|Average_Binder_Helix%"],
        "y": _descriptors_by_key["bindcraft|dssp|Average_Binder_BetaSheet%"],
    },
]



import subprocess
import os
import pandas as pd
import pytest
from ovo.core.utils.resources import RESOURCES_DIR
from ovo.core.utils.tests import get_test_scheduler, prepare_test_output_dir, assert_similar_sequence


def test_scaffold_hairpin():
    scheduler = get_test_scheduler()
    output_dir = prepare_test_output_dir(__file__)
    scheduler.run(
        "rfdiffusion-end-to-end",
        output_dir=output_dir,
        params=dict(
            max_memory="4G",
            batch_size=2,
            design_type="scaffold",
            rfdiffusion_input_pdb=RESOURCES_DIR / "examples/inputs/5ELI_A.pdb",
            rfdiffusion_num_designs=1,
            rfdiffusion_contig="A111-114/10/A117-119",
            rfdiffusion_run_parameters="diffuser.T=1 inference.deterministic=True",
            mpnn_run_parameters="--seed 42",
            refolding_tests="af2_model_1_ptm_ft_1rec,esmfold",
        ),
    )
    print("Saved to:", output_dir)
    subprocess.run(["ls", "-lR", output_dir.rstrip("/") + "/"], check=True)

    backbone_metrics = pd.read_csv(os.path.join(output_dir, "contig1_batch1/backbone_metrics.csv"))
    assert len(backbone_metrics) == 1
    first_bb = backbone_metrics.iloc[0]
    assert_similar_sequence(first_bb.pydssp_str, "--EEEEEE--EEEEEE-", 0.5)

    seq_composition = pd.read_csv(os.path.join(output_dir, "contig1_batch1/seq_composition.csv"))
    assert len(seq_composition) == 1
    first_seq = seq_composition.iloc[0]
    assert_similar_sequence(first_seq.sequence, "QSLHVNSSSGEAKVEAD", 0.5)

    initial_guess = pd.read_json(os.path.join(output_dir, "contig1_batch1/af2_model_1_ptm_ft_1rec.jsonl"), lines=True)
    assert len(initial_guess) == 1
    first_scores = initial_guess.iloc[0]
    assert first_scores["design_backbone_rmsd"] < 15.0
    assert first_scores["native_motif_rmsd"] < 20.0
    assert first_scores["plddt"] > 20.0

    esmfold = pd.read_json(os.path.join(output_dir, "contig1_batch1/esmfold.jsonl"), lines=True)
    assert len(esmfold) == 1
    first_esmfold = esmfold.iloc[0]
    assert first_esmfold["RMSD_backbone"] < 10.0
    assert first_esmfold["RMSD_all_atom"] < 20.0
    assert first_esmfold["pLDDT"] > 20.0



from abc import ABC, abstractmethod
from typing import Type, TypeVar, Sequence

import pandas as pd

from ovo.core.database.models import Base
from ovo.core.database.db_proxy import DBProxy

T = TypeVar("T", bound=Base)


class DBEngine(ABC):
    def init(self):
        pass

    @abstractmethod
    def save(self, obj: Base):
        raise NotImplementedError()

    @abstractmethod
    def save_all(self, objs: Sequence[Base]):
        raise NotImplementedError()

    @abstractmethod
    def remove(self, model: Type[T], *id_args, **kwargs):
        raise NotImplementedError()

    @abstractmethod
    def save_value(self, model: Type[T], column: str, value, *id_args, **kwargs):
        raise NotImplementedError()

    @abstractmethod
    def count(self, model: Type[T], field="id", **kwargs) -> Sequence[T]:
        raise NotImplementedError()

    @abstractmethod
    def count_distinct(self, model: Type[T], field="id", group_by=None, **kwargs) -> int:
        raise NotImplementedError()

    @abstractmethod
    def select(self, model: Type[T], limit: int = None, order_by=None, **kwargs) -> Sequence[T]:
        raise NotImplementedError()

    @abstractmethod
    def get(self, model: Type[T], *id_args, **kwargs) -> T:
        raise NotImplementedError()

    @abstractmethod
    def get_value(self, model: Type[T], column: str, *id_args, raw=False, **kwargs):
        raise NotImplementedError()

    @abstractmethod
    def select_values(self, model: Type[T], column: str, order_by=None, **kwargs) -> list:
        raise NotImplementedError()

    @abstractmethod
    def select_dict(self, model: Type[T], key_column: str, value_column: str, order_by=None, **kwargs) -> dict:
        raise NotImplementedError()

    @abstractmethod
    def select_unique_values(self, model: Type[T], column: str, **kwargs) -> set:
        raise NotImplementedError()

    @abstractmethod
    def select_descriptor_values(self, descriptor_key: str, design_ids: list[str]) -> pd.Series:
        raise NotImplementedError()

    @abstractmethod
    def select_wide_descriptor_table(self, design_ids: list[str], descriptor_keys: list[str], **kwargs) -> pd.DataFrame:
        raise NotImplementedError()

    def __getattr__(self, name: str) -> DBProxy:
        """Enable accessing models as attributes of the DBEngine instance.

        For example instead of db.select(Design) you can do db.Design.select().
        """
        for mapper in Base.registry.mappers:
            cls = mapper.class_
            if cls.__name__ == name:
                return cls
        raise AttributeError(f"DB model or function {name} not available")



from ovo.core.database import Threshold


def filter_designs_by_thresholds(
    all_design_ids: list[str],
    thresholds: dict[str, Threshold],
    # descriptor_key -> design_id -> value
    values: dict[str, dict[str, float | None]],
) -> tuple[list[str], dict[str, int]]:
    """Filter designs by thresholds (inclusive - <=, >=)

    :param all_design_ids: list of design ids
    :param thresholds: dictionary descriptor_key -> Threshold object
    :param values: dictionary descriptor_key -> descriptor values (index = design_id, value = descriptor value)

    :return: tuple:
        - list of design ids that pass the thresholds, in original order
        - dictionary descriptor_key -> number of designs that pass the threshold
    """
    filtered_design_ids = []
    num_accepted_by_descriptor = {}
    num_missing_by_descriptor = {}
    for design_id in all_design_ids:
        passes_all = True
        for descriptor_key, threshold in thresholds.items():
            if not threshold.enabled:
                continue
            if descriptor_key not in values:
                print("Available descriptor values:", values)
                raise ValueError(
                    f"Cannot filter by descriptor '{descriptor_key}', value is missing for all {len(all_design_ids):,} designs"
                )
            if design_id not in values[descriptor_key]:
                num_missing_by_descriptor[descriptor_key] = num_missing_by_descriptor.get(descriptor_key, 0) + 1
                passes_all = False
                continue
            passes = True
            if threshold.max_value is not None:
                passes &= values[descriptor_key][design_id] <= threshold.max_value
            if threshold.min_value is not None:
                passes &= values[descriptor_key][design_id] >= threshold.min_value
            passes_all &= passes
            if descriptor_key not in num_accepted_by_descriptor:
                num_accepted_by_descriptor[descriptor_key] = 0
            if passes:
                num_accepted_by_descriptor[descriptor_key] += 1
        if passes_all:
            filtered_design_ids.append(design_id)
    if num_missing_by_descriptor:
        print("Descriptors with missing descriptor values marked as NOT ACCEPTED:")
        for descriptor_key, num in num_missing_by_descriptor.items():
            print(descriptor_key, num)
    return filtered_design_ids, num_accepted_by_descriptor



import glob
import os
import time

import jsonschema
import typer
import sys
from ovo.cli.common import console, OVOCliError
from ovo.core.scheduler import NextflowScheduler
from ovo.core.utils.param_validation import validate_params
from ovo.core.scheduler.simple_queue_mixin import SimpleQueueMixin
import re

app = typer.Typer(pretty_exceptions_enable=False, help="Schedule and monitor OVO jobs")


@app.command(context_settings={"allow_extra_args": True, "ignore_unknown_options": True})
def run(
    pipeline_name: str = typer.Argument(default=None, help="Pipeline to submit"),
    output: str = typer.Argument(default=None, help="Output directory"),
    scheduler_key: str = typer.Option(None, "--scheduler", help="Scheduler key"),
    link: bool = typer.Option(default=None, help="Do not copy output files, create a symlink instead"),
    ctx: typer.Context = typer.Option(None),
    help: bool | None = typer.Option(
        None,
        "--help",
        "-h",
        help="Show pipeline-specific help",
    ),
):
    """Run OVO job synchronously and save output to directory (shorhand for submit --sync)"""
    from ovo import config, get_scheduler

    resolved_scheduler_key = scheduler_key or config.default_scheduler
    scheduler = get_scheduler(scheduler_key=resolved_scheduler_key)
    usage = "Usage: ovo scheduler run PIPELINE_NAME OUTPUT_DIR [OPTIONS] [PARAMS]"
    if help is True:
        if pipeline_name:
            schema = scheduler.get_param_schema(pipeline_name=pipeline_name)
            console.print(format_help(schema))
            return
        else:
            console.print(usage)
            return
    if pipeline_name and (not output or output.startswith("--")):
        schema = scheduler.get_param_schema(pipeline_name=pipeline_name)
        error = f"Missing OUTPUT. {usage}"
        raise OVOCliError(format_help(schema) + f"\n\n[red][bold]✘[/bold] {error}[/red]")
    return submit(
        pipeline_name=pipeline_name,
        output=output,
        link=link,
        scheduler_key=scheduler_key,
        sync=True,
        ctx=ctx,
    )


@app.command(context_settings={"allow_extra_args": True, "ignore_unknown_options": True})
def submit(
    pipeline_name: str = typer.Argument(default=None, help="Pipeline to submit"),
    output: str = typer.Option(None, "--output", help="Custom output directory (requires --sync)"),
    scheduler_key: str = typer.Option(None, "--scheduler", help="Scheduler key"),
    sync: bool = typer.Option(default=None, help="Run nextflow process synchronously, waiting for completion"),
    link: bool = typer.Option(
        default=None, help="Do not copy output files, create a symlink instead (when --output is used)"
    ),
    ctx: typer.Context = typer.Option(None),
    help: bool | None = typer.Option(
        None,
        "--help",
        "-h",
        help="Show pipeline-specific help",
    ),
):
    """Submit a job through OVO scheduler"""
    from ovo import config, get_scheduler, storage
    from ovo.core.utils.formatting import parse_args

    resolved_scheduler_key = scheduler_key or config.default_scheduler
    scheduler = get_scheduler(scheduler_key=resolved_scheduler_key)

    usage = "Usage: ovo scheduler submit PIPELINE_NAME [OPTIONS] [PARAMS]"
    if help is True:
        if pipeline_name:
            schema = scheduler.get_param_schema(pipeline_name=pipeline_name)
            console.print(format_help(schema))
            return
        else:
            console.print(usage)
            return

    if not pipeline_name:
        available_paths = scheduler.get_pipeline_names()
        sep = "\n "
        raise OVOCliError(f"Pipeline name is required. Available pipelines: \n{sep}{sep.join(available_paths)}")

    extra_args = ctx.args if hasattr(ctx, "args") else []
    params = parse_args(extra_args)

    schema = scheduler.get_param_schema(pipeline_name=pipeline_name)

    if not params:
        error = "Provide pipeline parameters using --my_param value --other_param value"
        raise OVOCliError(format_help(schema) + f"\n\n[red][bold]✘[/bold] {error}[/red]")

    if schema:
        try:
            validate_params(params, schema)
        except jsonschema.ValidationError as e:
            error = f"Input parameters are invalid: {e.message}"
            if e.path:
                error += f" ({'.'.join(map(str, e.path))})"
            raise OVOCliError(format_help(schema) + f"\n\n[red][bold]✘[/bold] {error}[/red]")

    if output:
        if not sync:
            raise OVOCliError("Output directory can only be set with --sync option")
        if os.path.exists(output):
            raise OVOCliError(f"Output directory already exists: '{output}'")

    console.print(f"Submitting [bold]{pipeline_name}[/bold] with [bold]{resolved_scheduler_key}[/bold] scheduler:")
    max_key_length = max(len(key) for key in params.keys())
    for key, value in params.items():
        console.print(f" {key.rjust(max_key_length)}: [green]{value}[/green]")
    console.print("")

    submitted = scheduler.submit(
        pipeline_name=pipeline_name,
        params=params,
        submission_args=dict(sync=sync, resolve_relative_paths=True),
    )
    if sync:
        process, job_id = submitted
        console.print("")
        if process.returncode != 0:
            console.print(f"[bold][red]✘[/red] Pipeline failed with exit code {process.returncode}[/bold]")
            sys.exit(4)
        console.print("[bold][green]✔[/green] Pipeline finished successfully[/bold]")
        scheduler_output_dir = scheduler.get_output_dir(job_id=job_id)
        if output:
            console.print(f"Job output directory: {scheduler_output_dir}")
            console.print(f"Saving output...")
            storage.sync_directory(scheduler_output_dir, output, link=link)
            console.print(f"[green]✔[/green] Output saved to: {os.path.abspath(output)}")
        else:
            console.print(f"[green]✔[/green] Output saved to: {scheduler_output_dir}")
    else:
        job_id = submitted
        console.print("")
        console.print(f"Running in background, use --sync to run synchronously and wait for completion.")
        console.print(f"[green]✔[/green] Job submitted: [bold]{job_id}[/bold]")
        scheduler_arg = f"--scheduler {scheduler_key} " if scheduler_key else ""
        console.print(f"[green]✔[/green] Output directory: [bold]{scheduler.get_output_dir(job_id)}[/bold]")
        console.print(f"[green]✔[/green] Check status with: [bold]ovo scheduler status {scheduler_arg}{job_id}[/bold]")
    return job_id


@app.command(context_settings={"allow_extra_args": True, "ignore_unknown_options": True})
def jupyter(
    dirs: list[str] = typer.Argument(help="Directory to run from. Will be mounted inside the container."),
    env: str = typer.Option(
        default=None,
        help="Name of conda environment or container to run in, for example rfdiffusion. Runs in current environment if not specified.",
    ),
    scheduler_key: str = typer.Option(None, "--scheduler", help="Scheduler key"),
    ip: str = typer.Option(default="0.0.0.0", help="Host to bind Jupyter server to"),
    port: int = typer.Option(default=8888, help="Port to bind Jupyter server to"),
    run_parameters: str = typer.Option(default="", help="Additional commandline parameters to pass to Jupyter server"),
    timeout: int = typer.Option(default=3600, help="Timeout in seconds to wait for Jupyter server to start"),
):
    """Run JupyterLab server in a selected environment as a job using OVO scheduler"""
    from ovo import config, get_scheduler

    params = {
        "dirs": ",".join([os.path.abspath(d) for d in dirs]),
        "env": env,
        "ip": ip,
        "port": port,
    }
    if run_parameters:
        params["run_parameters"] = run_parameters
    resolved_scheduler_key = scheduler_key or config.default_scheduler
    scheduler = get_scheduler(scheduler_key=resolved_scheduler_key)

    if not isinstance(scheduler, NextflowScheduler):
        raise NotImplementedError("Only NextflowScheduler currently supports Jupyter submission.")

    pipeline_name = "jupyterlab"
    console.print(
        f"Submitting [bold]{pipeline_name}[/bold] in [bold]{env or 'ovo'}[/bold] environment with [bold]{resolved_scheduler_key}[/bold] scheduler:"
    )
    max_key_length = max(len(key) for key in params.keys())
    for key, value in params.items():
        console.print(f" {key.rjust(max_key_length)}: [green]{value}[/green]")
    console.print("")

    job_id = scheduler.submit(
        pipeline_name=pipeline_name,
        params=params,
        submission_args=dict(sync=False, resolve_relative_paths=True),
    )
    # Wait for workdir
    workdir_prefix = None
    console.print("Waiting for Jupyter job to start...")
    for retry in range(timeout):
        time.sleep(1)
        if scheduler.get_result(job_id) is False:
            console.print(scheduler.get_log(job_id))
            console.print(f"[red][bold]✘[/bold] Jupyter job {job_id} failed[/red]")
            sys.exit(4)
        log = scheduler.get_log(job_id)
        # parse regex [5f/008512] Submitted process > JupyterLab
        if match := re.search(r"\[([^[]*)] Submitted process > JupyterLab", log or ""):
            workdir_prefix = match.group(1)
            break

    if not workdir_prefix:
        console.print(scheduler.get_log(job_id))
        console.print(
            f"[red][bold]✘[/bold] Failed to get Jupyter workdir for job {job_id}, please check job status[/red]"
        )
        sys.exit(4)

    workdir_glob = os.path.join(scheduler.workdir, f"work/{workdir_prefix}*")
    workdir = glob.glob(workdir_glob)
    if not workdir:
        console.print(f"[red][bold]✘[/bold] Failed to find Jupyter workdir: {workdir_glob}[/red]")
        sys.exit(4)
    workdir = workdir[0]

    console.print(f"Checking workdir {workdir} for Jupyter URL...")
    jupyter_url_lines = None
    hostname = None
    installing_jupyter = None
    log = None
    for retry in range(timeout):
        time.sleep(1)
        log_path = os.path.join(workdir, ".command.log")
        if scheduler.get_result(job_id) is False:
            if os.path.exists(log_path):
                with open(log_path) as f:
                    console.print(f.read())
            else:
                console.print(scheduler.get_log(job_id))
            console.print(f"[red][bold]✘[/bold] Jupyter job {job_id} failed[/red]")
            sys.exit(4)
        if not os.path.exists(log_path):
            continue
        with open(log_path) as f:
            log = f.read()
        if "Jupyter Lab not available" in log:
            if not installing_jupyter:
                console.print("Jupyter Lab not available in the environment, installing...")
                installing_jupyter = True
        jupyter_url_lines = [
            line
            for line in log.splitlines()
            if ("http://" in line or "https://" in line) and "token=" in line and "ServerApp" not in line
        ]
        if jupyter_url_lines:
            hostname_lines = [line for line in log.splitlines() if "JUPYTER_HOSTNAME:" in line]
            if hostname_lines:
                hostname = hostname_lines[-1].split("JUPYTER_HOSTNAME:")[-1].strip()
            break

    if not jupyter_url_lines:
        console.print(log or scheduler.get_log(job_id))
        console.print(f"[red][bold]✘[/bold] Failed to get Jupyter URL for job {job_id}, please check job status[/red]")
        sys.exit(4)

    console.print(f"\n[green]✔[/green] Jupyter [bold]{env or 'ovo'}[/bold] environment is running at:")

    different_port = False
    for line in jupyter_url_lines:
        console.print(line)
        if str(port) not in line:
            different_port = True
    if different_port:
        console.print(
            f"\n[bold][red]Note![/red][/bold] JupyterLab is running on a different port! The requested port {port} was occupied.\n"
        )

    console.print("Tips:")
    console.print(" - Cmd/Ctrl + Double click the URL to open the link in your browser.")
    if hostname:
        console.print(
            f" - If the host is not directly accessible, use SSH tunneling to forward your local port {port} to remote {hostname}:{port}"
        )
    console.print(
        f" - Remember to stop the job when you're done using 'ovo scheduler cancel {job_id}' or directly from JupyterLab using Server -> Shutdown"
    )


def format_help(schema):
    from ovo import config

    params_str = "Available schedulers:\n\n"
    for scheduler_key, scheduler_config in config.schedulers.items():
        params_str += (
            f" --scheduler {scheduler_key} ({scheduler_config.name}, args: {scheduler_config.submission_args})\n"
        )
    params_str += "\n"
    if schema:
        sep = "\n  "
        formatted_params = {
            key: f"--{key}: {value['description']}" + (f" (default {value['default']})" if value.get("default") else "")
            if value.get("description")
            else f"--{key}"
            for key, value in schema.get("properties", {}).items()
        }
        required_params = [formatted_params.get(key, f"--{key}") for key in schema.get("required", [])]
        if required_params:
            params_str += f"Required parameters:\n{sep}{sep.join(required_params)}\n\n"
        optional_params = [
            formatted_params[key]
            for key in schema.get("properties", {}).keys()
            if key not in schema.get("required", [])
        ]
        params_str += f"Optional parameters:\n{sep}{sep.join(optional_params) or 'None'}"
    return params_str.strip()


@app.command()
def status(
    job_id: str = typer.Argument(help="Job ID"),
    scheduler_key: str = typer.Option(None, "--scheduler", help="Scheduler key"),
):
    """Get job status"""
    from ovo import config, get_scheduler

    resolved_scheduler_key = scheduler_key or config.default_scheduler
    scheduler = get_scheduler(scheduler_key=resolved_scheduler_key)

    job_result = scheduler.get_result(job_id=job_id)
    if job_result is None:
        console.print(f"[bold]Job {job_id} is running[/bold]")
        sys.exit(3)
    elif job_result is False:
        console.print(f"[bold][red]✘[/red] Job {job_id} failed[/bold]")
        console.print(scheduler.get_log(job_id=job_id))
        sys.exit(4)
    else:
        console.print(f"[bold][green]✔[/green] Job {job_id} completed successfully[/bold]")
        console.print(f"Output directory: {scheduler.get_output_dir(job_id=job_id)}")


@app.command()
def cancel(
    job_id: str = typer.Argument(help="Job ID"),
    scheduler_key: str = typer.Option(None, "--scheduler", help="Scheduler key"),
):
    """Cancel job execution"""
    from ovo import config, get_scheduler

    resolved_scheduler_key = scheduler_key or config.default_scheduler
    scheduler = get_scheduler(scheduler_key=resolved_scheduler_key)

    console.print(f"Cancelling job [bold]{job_id}[/bold] with scheduler [bold]{resolved_scheduler_key}[/bold]")
    scheduler.cancel(job_id=job_id)
    console.print(f"[green]✔[/green] Job cancelled")


@app.command()
def download(
    job_id: str = typer.Argument(help="Job ID"),
    output: str = typer.Argument(help="Output directory"),
    scheduler_key: str = typer.Option(None, "--scheduler", help="Scheduler key"),
):
    """Download job output files into a directory"""
    from ovo import config, get_scheduler, storage

    resolved_scheduler_key = scheduler_key or config.default_scheduler

    status(job_id=job_id, scheduler_key=resolved_scheduler_key)

    scheduler = get_scheduler(scheduler_key=resolved_scheduler_key)
    scheduler_output_dir = scheduler.get_output_dir(job_id=job_id)

    console.print(f"Job output directory: {scheduler_output_dir}")
    console.print(f"Saving output to: [bold]{output}[/bold]")
    storage.sync_directory(scheduler_output_dir, output, link=True)
    console.print(f"Output saved to: {os.path.abspath(output)}")


@app.command()
def worker(
    num_workers: int = typer.Option(1, "-n", "--workers", help="Number of worker threads"),
    connect: bool = typer.Option(
        False,
        "--connect",
        help="Connect to existing queue server instead of starting a new one. Enables running workers with different ENV vars such as CUDA_VISIBLE_DEVICES.",
    ),
    scheduler_key: str = typer.Option(None, "--scheduler", help="Scheduler key to use"),
    host: str = typer.Option(None, "--host", help="Run server on custom host (override config)"),
    port: int = typer.Option(None, "--port", help="Run server on custom port (override config)"),
):
    """Start queue and worker server to process jobs of a scheduler specified with queue submission arg in OVO config"""
    from ovo import config, get_scheduler
    import threading

    resolved_scheduler_key = scheduler_key or config.default_scheduler

    scheduler: SimpleQueueMixin = get_scheduler(scheduler_key=resolved_scheduler_key)

    if connect:
        # Connect mode: use already running queue
        queue = scheduler.connect_to_queue(host=host, port=port)
    else:
        # Start server in the background
        queue, serve_forever = scheduler.create_queue_server(host=host, port=port)
        threading.Thread(target=serve_forever, daemon=True).start()

    # Start workers
    threads = []
    for i in range(num_workers):
        t = threading.Thread(target=scheduler.queue_worker, args=(queue, i + 1))
        t.start()
        threads.append(t)

    # Keep main thread alive if connecting, or join worker threads
    try:
        for t in threads:
            t.join()
    except KeyboardInterrupt:
        for n in range(num_workers + 100):
            # Send shutdown signal to workers, up to 10 extra in case some were executed independently
            queue.put(None)
        # TODO this actually doesn't seem to work,
        #  workers exit immediately even if nextflow run is still in progress
        print("\nWaiting for workers to finish their task, use Ctrl+C again to force shutdown...")
        for t in threads:
            t.join()


if __name__ == "__main__":
    app()



import json
import os
import streamlit.components.v1 as components

parent_dir = os.path.dirname(os.path.abspath(__file__))
build_dir = os.path.join(parent_dir, "frontend/build")
_component_func = components.declare_component("contigs_organizer", path=build_dir)


# Create a wrapper function for the component. This is an optional
# best practice to validate input and add documentation comments.
def contigs_organizer(contigs: str, pdb: str, colors: dict[str, str], key: str):
    """Create a new instance of "contigs_organizer".

    Parameters
    ----------
    contigs: str
        Original contigs definition
    pdb: str
        PDB of the relevant structure
    key: str
        A key that uniquely identifies this component to avoid re-mounting on changed parameters.
    colors: dict[str, str]
        Dictionary mapping segment strings (A123-456) to colors (#00ff00)

    Returns
    -------
    A JSON information about the organized contigs.

    """
    component_value = _component_func(contigs=contigs, pdb=pdb, colors=json.dumps(colors), key=key)

    return component_value



import json
import os
from io import StringIO
from typing import Callable

import pandas as pd

from ovo import storage, Pool, Round, db, get_scheduler
from ovo.core.database import DesignJob, DesignSpec, Design, Base
from ovo.core.database.models_bindcraft import BindCraftBinderDesignWorkflow
from ovo.app.utils.bindcraft_utils import load_json_from_file, merge_dictionaries
from ovo.core.logic.descriptor_logic import read_descriptor_file_values, save_descriptor_job_for_design_job


def prepare_bindcraft_params(workflow: BindCraftBinderDesignWorkflow, workdir: str) -> dict:
    input_dict = {
        "design_path": "output",
        "starting_pdb": "target.pdb",
        "binder_name": "design",
        "chains": workflow.bindcraft_params.target_chains,
        "target_hotspot_residues": workflow.bindcraft_params.hotspots,
        "lengths": [int(v) for v in workflow.bindcraft_params.binder_length.split(",")],
        "number_of_final_designs": workflow.bindcraft_params.number_of_final_designs,
    }

    settings_advanced, settings_filters = workflow.get_settings_paths()

    merged_advanced_dict = merge_dictionaries(
        load_json_from_file(settings_advanced), workflow.bindcraft_params.custom_advanced_settings
    )
    merged_filter_dict = merge_dictionaries(
        load_json_from_file(settings_filters), workflow.bindcraft_params.custom_filter_settings
    )

    return {
        "time_limit_seconds": workflow.bindcraft_params.time_limit_hours * 3600,
        "input_pdb": storage.prepare_workflow_input(workflow.bindcraft_params.input_pdb_path, workdir=workdir),
        "input_json_path": storage.prepare_workflow_input(
            "input.json",
            workdir=workdir,
            input_bytes=json.dumps(input_dict).encode("utf-8"),
        ),
        "settings_advanced": storage.prepare_workflow_input(
            "settings_advanced.json",
            workdir=workdir,
            input_bytes=json.dumps(merged_advanced_dict, indent=2).encode("utf-8"),
        ),
        "settings_filters": storage.prepare_workflow_input(
            "settings_filters.json",
            workdir=workdir,
            input_bytes=json.dumps(merged_filter_dict, indent=2).encode("utf-8"),
        ),
        "num_replicas": workflow.bindcraft_params.num_replicas,
    }


def process_workflow_results(job: DesignJob, callback: Callable = None) -> list[Base]:
    pool = db.get(Pool, design_job_id=job.id)
    project_round = db.get(Round, id=pool.round_id)
    scheduler = get_scheduler(job.scheduler_key)

    # this is where result files will be stored in our storage
    # make sure to remove trailing slash otherwise S3 will keep two slashes in the path
    destination_dir = os.path.join("project", project_round.project_id, "pools", pool.id, "designs").rstrip("/")

    source_output_path = scheduler.get_output_dir(job.job_id)

    num_replicas = job.workflow.bindcraft_params.num_replicas

    designs = []
    final_rows = []
    design_id_mapping = {}
    for replica in range(1, num_replicas + 1):
        batch_dir = f"batch{replica}"
        accepted_dir = os.path.join(source_output_path, f"{batch_dir}/bindcraft/Accepted")
        accepted_filenames = [filename for filename in storage.list_dir(accepted_dir) if filename.endswith(".pdb")]
        accepted_designs = [filename.split("_model")[0] for filename in accepted_filenames]

        final_designs = pd.read_csv(
            StringIO(
                storage.read_file_str(os.path.join(source_output_path, f"{batch_dir}/bindcraft/final_design_stats.csv"))
            ),
        ).drop(columns=["Rank"])
        if missing_designs := set(final_designs["Design"]).difference(accepted_designs):
            raise FileNotFoundError(
                f"Unexpected error: Missing files for accepted designs: {missing_designs}, "
                f"found only {accepted_designs} in {accepted_dir}"
            )

        rejected_dir = os.path.join(source_output_path, f"{batch_dir}/bindcraft/Rejected")
        rejected_filenames = [filename for filename in storage.list_dir(rejected_dir) if filename.endswith(".pdb")]

        design_df = pd.read_csv(
            StringIO(
                storage.read_file_str(os.path.join(source_output_path, f"{batch_dir}/bindcraft/mpnn_design_stats.csv"))
            ),
        )
        design_df = design_df.sort_values("Average_i_pTM", ascending=False)

        # check the ranking of the designs and copy them with new ranked IDs to the folder
        rank = 1
        rejected = 1
        for i, (_, row) in enumerate(design_df.iterrows()):
            model_filenames = [
                filename
                for filename in accepted_filenames + rejected_filenames
                if filename.startswith(row["Design"] + "_model")
            ]
            if not model_filenames:
                raise ValueError(
                    f"Unexpected error: Missing accepted or rejected PDB model files for design {row['Design']}"
                )

            for filename in model_filenames:
                design, model = filename.removesuffix(".pdb").split("_model")

                if callback:
                    callback(
                        value=(i + 1) / len(design_df),
                        text=f"Downloading design {design}",
                    )

                accepted = filename in accepted_filenames
                id_prefix = f"ovo_{pool.id}"
                if num_replicas > 1:
                    replica_str = str(replica).zfill(len(str(num_replicas)))
                    id_prefix += f"_batch{replica_str}"

                if accepted:
                    structure_path = os.path.join(accepted_dir, filename)
                    rank_str = str(rank).zfill(max(2, len(str(rank))))
                    design_id = f"{id_prefix}_rank{rank_str}_bindcraft"
                    rank += 1
                else:
                    structure_path = os.path.join(rejected_dir, filename)
                    rejected_str = str(rejected).zfill(max(2, len(str(rejected))))
                    design_id = f"{id_prefix}_rejected{rejected_str}_bindcraft"
                    rejected += 1
                design_id_mapping[design_id] = design_id
                final_rows.append(
                    {
                        "ID": design_id,
                        "Rank": rank if accepted else None,
                        "Model": model,
                        **row[final_designs.columns].to_dict(),
                    }
                )
                design = Design(
                    id=design_id,
                    pool_id=pool.id,
                    accepted=accepted,
                    structure_path=storage.store_file_path(
                        structure_path,
                        os.path.join(destination_dir, f"{design_id}.pdb"),
                    ),
                )
                design.spec = DesignSpec.from_pdb_str(
                    pdb_data=storage.read_file_str(design.structure_path), chains=["B"]
                )
                designs.append(design)

    if not final_rows:
        # Case when no trajectories proceeded to filtering stage
        job.job_result = False
        job.warnings.append(
            f"No designs found! Please use a higher time limit. You can inspect discarded trajectories in the output path: {source_output_path}"
        )
        # Job will be saved by caller
        return

    # ID, Rank, Model, Average_i_pTM, ...
    final_df = pd.DataFrame(final_rows).set_index("ID")

    # IMPORTANT
    # We convert pLDDT to 0-100 scale and PAE to 0-31 scale to be consistent with ColabDesign
    for column in final_df.columns:
        if "_plddt" in column.lower():
            final_df[column] *= 100
        if "_pae" in column.lower():
            final_df[column] *= 31

    descriptor_job = save_descriptor_job_for_design_job(
        design_job=job,
        project_id=project_round.project_id,
        # TODO currently binder chain is B, this might be changed later to chain A
        chains=["B"],
        design_ids=list(design_id_mapping.keys()),
    )
    descriptor_values = read_descriptor_file_values(
        descriptor_job=descriptor_job,
        design_id_mapping=design_id_mapping,
        descriptor_tables={
            "bindcraft|af2": final_df,
            "bindcraft|mpnn": final_df,
            "bindcraft|interface": final_df,
            "bindcraft|dssp": final_df,
        },
    )

    return designs + descriptor_values



import streamlit as st
import json

from ovo.app.components.history_components import history_dropdown_component
from ovo.app.components.input_components import pdb_input_component, sequence_selection_fragment, initialize_workflow
from ovo.app.components.molstar_custom_component import molstar_custom_component, StructureVisualization
from ovo.app.components.navigation import show_prev_next_sections
from ovo.app.components.submission_components import (
    pool_submission_inputs,
    show_bindcraft_advanced_settings,
    review_workflow_submission,
)

from ovo.app.pages import jobs_page, designs_page
from ovo.app.utils.page_init import initialize_page
from ovo import storage
from ovo.core.database.models_bindcraft import BindCraftBinderDesignWorkflow
from ovo.core.utils.formatting import get_hashed_path_for_bytes
from ovo.core.utils.residue_selection import get_chains_and_contigs


@st.fragment
def intro_step():
    # Initialize the workflow object in session state
    initialize_workflow(__file__, BindCraftBinderDesignWorkflow.name)

    with st.container(width=850):
        st.markdown(
            f"""
Miniprotein binder design using BindCraft.<sup>[1]</sup> BindCraft is a binder design pipeline using AlphaFold2<sup>[2]</sup>
backpropagation via ColabDesign<sup>[3]</sup>, followed by ProteinMPNN<sup>[4]</sup> surface sequence design, and PyRosetta<sup>[5]</sup> scoring.

Use case: Designing linear peptide or miniprotein binders to a target protein.
           
- **Input**
    - Target protein structure (PDB file or UniProt ID)
- **Main parameters**
    - Target chain
    - Binder length range (peptide to miniprotein)
    - Number of final *accepted* designs to generate, and a time limit
    - (Optional) Hotspot residues to select a preferred binding site
    - (Optional) Design protocol (Default, Beta-sheet, Peptide)
    - (Optional) Filtering settings
- **Output**
    - Designs — binder sequences and structures in complex with target protein
    - Design descriptors from BindCraft
        - Refolding metrics
        - PyRosetta descriptors
        - Secondary structure

### Features
        
After inputting your target structure, you can choose the desired binding site by selecting "hotspot" residues, 
or you can leave it blank to allow binding anywhere on the target surface. Next, you can specify optional advanced settings,
including the design protocol and filtering settings before submitting your job. 

Once the Bindcraft workflow has finished, you will see all generated designs (rejected and accepted by final filters) and their descriptors 
in {jobs_page.title} results page. 
Designs that pass BindCraft filters will be marked as *accepted designs* and shown in {designs_page.title} page, where
you can run ProteinQC, computing descriptors of predicted solubility, electrostatic and hydrophobic patch area, and more.    

The BindCraft workflow has three parts — a design phase, sequence redesign phase, and a filtering phase. 

The **design phase** iteratively refines a binder, starting from a random sequence, with AlphaFold2 (the multimer model) optimizing the
AlphaFold2 confidence metrics, binder-target contacts, and other differentiable objectives. 
By default, the 4-stage protocol is used, which has been validated by
the authors across several targets. Only if BindCraft fails to produce successful designs, the authors suggest tweaking
the design settings (see BindCraft supplement for details). With the exception of peptide binders (8–25
residues): The authors recommend the peptide-specific design (3-stage protocol) and filtering settings. 
                
The **sequence redesign phase** resamples the sequence of the non-interface residues using SolubleMPNN (by default; or
ProteinMPNN). It has been shown to improve expression and solubility compared to using the AlphaFold2-hallucinated
sequences alone.

The **filtering phase** uses the AlphaFold2 monomer model (by default), which has not seen multimer interfaces in its training
data, and checks that "the interface is sufficiently well-defined to be detected even by a model that has never been
exposed to multimeric complexes." Besides filtering for high confidence metrics, it verifies the RMSD between the
original hallucinated binder pose with AF2 multimer and the MPNN-redesigned AF2-monomer repredicted binder (AF2 Hotspot
RMSD) is sufficiently low. The default filters also filter out unstructured binders (by secondary structure), enforce
a number of interface residue contacts, etc. The designs that pass BindCraft filters are automatically marked as accepted.

BindCraft refolding metrics are similar to those in the RFdiffusion workflow, with the addition of binder metrics
computed using AF2 prediction of the binder alone — e.g. AF2 Binder pLDDT, or AF2 Binder-aligned Binder RMSD. Here the binder-only
structural model is compared to the binder in the design trajectory, the complex. (Note that this definition is different to
the metric with the same name in RFdiffusion workflow.) 
        
### Limitations

- ⚠️ **BindCraft is shown to be more compute efficient than RFdiffusion** with respect to the number of accepted designs, 
  but it is more memory intensive due to using multiple AlphaFold2 model weights concurrently. If you encounter out-of-memory errors,
  please try trimming the target chain or configuring your scheduler to use a larger GPU instance.
- ⚠️ **Discarded trajectories are not saved in OVO storage**. Please inspect the OVO workdir of your job if you need them (you will need access to the filesystem).
- ⚠️ We preserve both the rejected and accepted designs for your inspection, 
  but unlike in our RFdiffusion workflow, **BindCraft filter thresholds cannot be changed after job submission**.

### References 
1. [BindCraft](https://www.nature.com/articles/s41586-025-09429-6), [BindCraft GitHub](https://github.com/martinpacesa/BindCraft)
2. [AlphaFold2](https://www.nature.com/articles/s41586-021-03819-2)
3. [ColabDesign](https://github.com/sokrypton/ColabDesign)
4. [ProteinMPNN](https://www.science.org/doi/10.1126/science.add2187)
5. [PyRosetta](https://github.com/RosettaCommons/PyRosetta.notebooks)
""",
            unsafe_allow_html=True,
        )


@st.fragment
def input_step():
    workflow: BindCraftBinderDesignWorkflow = st.session_state.workflows[__file__]

    left, _, right = st.columns([4, 1, 3])
    with left:
        st.subheader("Input structure")

        if not workflow.bindcraft_params.input_pdb_path:
            if new_pdb_input := pdb_input_component(""):
                input_name, pdb_input_bytes = new_pdb_input
                workflow.selected_segments = None
                workflow.input_name = input_name
                workflow.bindcraft_params.hotspots = None
                workflow.bindcraft_params.input_pdb_path = storage.store_file_str(
                    pdb_input_bytes.decode(),
                    f"project/{st.session_state.project.id}/inputs/{get_hashed_path_for_bytes(pdb_input_bytes)}/{workflow.input_name}.pdb",
                    overwrite=False,
                )
                workflow.chains = get_chains_and_contigs(pdb_input_bytes.decode())
                st.rerun()
        elif st.button("Change input"):
            workflow.input_name = None
            workflow.set_selected_segments(None)
            workflow.bindcraft_params.input_pdb_path = None
            st.rerun()

    with right:
        history_dropdown_component(__file__, workflow_name=BindCraftBinderDesignWorkflow.name)

    if not workflow.bindcraft_params.input_pdb_path:
        # No input structure provided yet, stop here
        return

    st.subheader(workflow.input_name)

    molstar_custom_component(
        structures=[
            StructureVisualization(
                pdb=storage.read_file_str(workflow.bindcraft_params.input_pdb_path), color="chain-id"
            )
        ],
        key="input_structure",
    )


@st.fragment
def hotspots_step():
    workflow: BindCraftBinderDesignWorkflow = st.session_state.workflows[__file__]

    st.subheader("Select binding hotspots")

    if not workflow.bindcraft_params.input_pdb_path:
        st.error("Please provide an input structure in the input structure step.")
        return

    sequence_selection_fragment(__file__, workflow.input_name, color="hydrophobicity", write_segments=False)


@st.fragment()
def settings_step():
    st.subheader("Settings")

    workflow: BindCraftBinderDesignWorkflow = st.session_state.workflows[__file__]

    if not workflow.bindcraft_params.input_pdb_path:
        st.error("Please provide an input structure in the input structure step.")
        return

    pool_submission_inputs(__file__)

    with st.columns([1, 2])[0]:
        if workflow.bindcraft_params.target_chains is None and len(workflow.chains) == 1:
            # If there is only one chain, set it as the default target chain
            workflow.bindcraft_params.target_chains = list(workflow.chains.keys())[0]

        workflow.bindcraft_params.target_chains = st.text_input(
            "Target chain(s)",
            value=workflow.bindcraft_params.target_chains,
            key="target_chains",
            help="Which chains of your PDB to target? Can be one or multiple, in a comma-separated format. Other chains will be ignored during design.",
        )
        if len(workflow.chains) > 1:
            st.write(f"*Available chains: {','.join(workflow.chains.keys())}*")

        workflow.bindcraft_params.binder_length = st.text_input(
            "Binder length",
            value=workflow.bindcraft_params.binder_length,
            key="binder_length",
            help="Comma-separated range of minimum and maximum number of residues in designed binder. Length of each design will be uniformly sampled from this range.",
        )

        workflow.bindcraft_params.hotspots = st.text_input(
            "Hotspot residues (optional)", value=workflow.bindcraft_params.hotspots, key="hotspots"
        )

        workflow.bindcraft_params.num_replicas = st.number_input(
            "Number of parallel replicas (GPU jobs)",
            value=workflow.bindcraft_params.num_replicas,
            key="num_replicas",
            min_value=1,
            max_value=20,
        )

        workflow.bindcraft_params.number_of_final_designs = st.number_input(
            "Desired number of final designs passing all filters (per replica)",
            value=workflow.bindcraft_params.number_of_final_designs,
            key="number_of_final_designs",
            min_value=0,
        )

        workflow.bindcraft_params.time_limit_hours = st.number_input(
            "Time limit (hours)",
            value=workflow.bindcraft_params.time_limit_hours,
            key="time_limit_hours",
            max_value=24 * 14,
        )
        summary = (
            f"**Summary:** Will run {workflow.bindcraft_params.num_replicas} parallel replicas, each stopping when it reaches "
            f"the time limit of {workflow.bindcraft_params.time_limit_hours} hours or when it generates "
            f"{workflow.bindcraft_params.number_of_final_designs} accepted designs. "
        )
        if workflow.bindcraft_params.num_replicas > 1:
            summary += f"Total number of accepted designs can be up to {workflow.bindcraft_params.num_replicas * workflow.bindcraft_params.number_of_final_designs}."
        st.write(summary)

        if workflow.bindcraft_params.num_replicas > 1 and workflow.bindcraft_params.time_limit_hours:
            st.caption(
                f"Each replica will run up to {workflow.bindcraft_params.time_limit_hours} hours. Total runtime may be longer in case when all replicas cannot run in parallel."
            )

        # Dropdowns for user selections
        workflow.bindcraft_params.design_protocol = st.selectbox(
            "Which binder design protocol to run?",
            options=["Default", "Beta-sheet", "Peptide"],
            help=(
                '"Default" is recommended. '
                '"Beta-sheet" promotes the design of more beta-sheeted proteins, but requires more sampling. '
                '"Peptide" is optimized for helical peptide binders.'
            ),
        )

        workflow.bindcraft_params.prediction_protocol = st.selectbox(
            "What prediction protocol to use?",
            options=["Default", "HardTarget"],
            help=(
                '"Default" performs single-sequence prediction of the binder. '
                '"HardTarget" uses an initial guess to improve complex prediction for difficult targets, but may introduce bias.'
            ),
        )

        workflow.bindcraft_params.interface_protocol = st.selectbox(
            "What interface design method to use?",
            options=["AlphaFold2", "MPNN"],
            help=(
                '"AlphaFold2" is the default; the interface is generated by AlphaFold2. '
                '"MPNN" uses soluble MPNN to optimize the interface.'
            ),
        )

        workflow.bindcraft_params.template_protocol = st.selectbox(
            "What target template protocol to use?",
            options=["Default", "Masked"],
            help=(
                '"Default" allows for limited flexibility. '
                '"Masked" allows for greater target flexibility on both sidechain and backbone levels.'
            ),
        )

        types = ["Default", "Peptide", "Relaxed", "Peptide_Relaxed", "None"]

        workflow.bindcraft_params.filter_type = st.selectbox(
            "Filtering",
            key="filtering",
            options=types,
            index=(
                types.index(workflow.bindcraft_params.filter_type)
                if workflow.bindcraft_params.filter_type in types
                else 0
            ),
        )

    show_bindcraft_advanced_settings(workflow)


@st.fragment
def review_step():
    st.subheader("Review settings")

    review_workflow_submission(__file__)


if __name__ == "__main__":
    initialize_page(page_title="BindCraft binder design")

    show_prev_next_sections(
        key=__file__,
        title="️⚒️ BindCraft binder design",
        sections={
            "intro": intro_step,
            "input": input_step,
            "hotspots": hotspots_step,
            "settings": settings_step,
            "review": review_step,
        },
    )



import sys
import traceback

from copy import deepcopy
from datetime import datetime
from typing import Callable

import pandas as pd
from humanize import naturaltime
from sqlalchemy.orm.attributes import flag_modified

from ovo import db, config
from ovo import get_scheduler
from ovo.core.auth import get_username
from ovo.core.database.models import Design, Threshold, DescriptorValue, Round, DesignJob, Base, DesignWorkflow
from ovo.core.database.models import Pool, Workflow
from ovo.core.logic.filtering_logic import filter_designs_by_thresholds
from ovo.core.logic.job_logic import update_job_status, format_job_duration


def get_design_jobs_table(
    project_id: str = None, round_ids: list[str] = None, update=True, **pool_filters
) -> pd.DataFrame:
    """Get pandas DataFrame with job and pool metadata and workflow parameter information."""
    id_filters = {}
    if project_id is not None:
        project_round_ids = db.select_unique_values(Round, "id", project_id=project_id)
        if round_ids is not None:
            round_ids = set(round_ids).intersection(set(project_round_ids))
        else:
            round_ids = project_round_ids
    if round_ids is not None:
        if isinstance(round_ids, str):
            round_ids = [round_ids]
        id_filters["round_id__in"] = round_ids

    pools = db.select(Pool, **id_filters, **pool_filters)
    pools_by_job_id = {p.design_job_id: p for p in pools if p.design_job_id is not None}
    jobs = db.select(DesignJob, order_by="-created_date_utc", id__in=pools_by_job_id.keys())

    # Update job.job_result. If job failed, it will be saved to the DB.
    # If job was successful, it's only updated in the job object, we will save it to DB later (after the designs have been processed)
    for job in jobs:
        if job.job_result is None and update:
            update_job_status(job)

    accepted_by_pool = db.count_distinct(Design, group_by="pool_id", pool_id__in=[p.id for p in pools], accepted=True)
    total_by_pool = db.count_distinct(Design, group_by="pool_id", pool_id__in=[p.id for p in pools])

    rows = []
    for job in jobs:
        pool = pools_by_job_id.get(job.id)
        rows.append(
            {
                ("Pool", "id"): pool.id,
                ("Pool", "name"): pool.name,
                ("Pool", "description"): pool.description,
                ("Job", "status"): format_pool_status(job, pool.processed, update_status=update),
                ("Job", "duration"): format_job_duration(job),
                ("Job", "created"): naturaltime(job.created_date_utc, when=datetime.utcnow()),
                ("Designs", "accepted"): accepted_by_pool.get(pool.id, 0) if pool.processed else None,
                ("Designs", "total"): total_by_pool.get(pool.id, 0) if pool.processed else None,
            }
        )

    if not rows:
        return pd.DataFrame()

    pools_table = pd.DataFrame(rows)
    pools_table.columns = pd.MultiIndex.from_tuples([c for c in pools_table.columns])

    workflows_table = get_workflows_table(jobs=jobs)

    pools_table = pd.concat([pools_table, workflows_table], axis=1)

    return pools_table.set_index([("Pool", "id")])


def get_pools_table(project_id: str = None, round_ids: list[str] = None):
    id_filters = {}
    if project_id is not None:
        project_round_ids = db.select_unique_values(Round, "id", project_id=project_id)
        if round_ids is not None:
            round_ids = set(round_ids).intersection(set(project_round_ids))
        else:
            round_ids = project_round_ids
    if round_ids is not None:
        if isinstance(round_ids, str):
            round_ids = [round_ids]
        id_filters["round_id__in"] = round_ids

    pools = db.select(Pool, **id_filters, order_by="-created_date_utc")

    accepted_by_pool = db.count_distinct(Design, group_by="pool_id", pool_id__in=[p.id for p in pools], accepted=True)
    total_by_pool = db.count_distinct(Design, group_by="pool_id", pool_id__in=[p.id for p in pools])
    failed_job_ids = db.select_unique_values(
        DesignJob, "id", job_result=False, id__in=[p.design_job_id for p in pools if p.design_job_id]
    )

    df = pd.DataFrame(
        [
            {
                "ID": pool.id,
                "Name": pool.name,
                "Description": pool.description,
                "Created": naturaltime(pool.created_date_utc, when=datetime.utcnow()),
                "Accepted Designs": accepted_by_pool.get(pool.id, 0) if pool.processed else "Not processed yet",
                "Total Designs": total_by_pool.get(pool.id, 0) if pool.processed else None,
                "Job ID": pool.design_job_id,
            }
            for pool in pools
            if (not pool.design_job_id or pool.design_job_id not in failed_job_ids)
        ]
    )
    if len(round_ids) > 1:
        round_names = db.select_dict(Round, "id", "name", id__in=round_ids)
        df.insert(0, "Round", [round_names.get(pool.round_id) for pool in pools])

    return df


def format_pool_status(job: DesignJob, processed: bool, update_status: bool = True):
    job_result = job.job_result
    if processed == True and job_result is None:
        return None
    if job_result == True:
        return "✅ Ready" if not processed else "Done"
    elif job_result == False:
        return "Failed"
    if update_status:
        # Get status label from scheduler
        try:
            scheduler = get_scheduler(job.scheduler_key)
        except:
            traceback.print_exc()
            return "Unknown Scheduler"
        return f"⏳{scheduler.get_status_label(job.job_id)}"
    return "In progress"


def get_workflows_table(jobs: list[DesignJob]):
    """Get pandas DataFrame with workflow parameter information for each pool."""
    table = pd.DataFrame([job.workflow.get_table_row() for job in jobs])
    # If there are no pools with a workflow, create empty multiindex manually
    if table.columns.empty and not table.empty:
        table.columns = pd.MultiIndex.from_tuples([])
    return table


def submit_design_workflow(
    workflow: Workflow,
    scheduler_key: str,
    round_id: str,
    pool_name: str,
    pool_description: str,
    return_existing: bool = True,
    pipeline_name: str = None,
) -> tuple[DesignJob, Pool]:
    """Submit a design workflow to the scheduler and create a Pool and DesignJob in the DB.

    :param workflow: Workflow object to submit
    :param scheduler_key: Key of the scheduler to use
    :param round_id: ID of the Round to associate the Pool with
    :param pool_name: Name of the Pool to create
    :param pool_description: Description of the Pool to create
    :param return_existing: If a Pool with the same name and parameters already exists in this round, return it instead of raising an error
    :param pipeline_name: Override the pipeline name to submit, e.g. ovo.rfdiffusion-end-to-end or a github url with @version
    :return: Tuple of (DesignJob, Pool)
    """
    scheduler = get_scheduler(scheduler_key)

    if existing_pools := db.select(Pool, name=pool_name, round_id=round_id):
        if not return_existing:
            raise ValueError(f"Pool with name '{pool_name}' already exists in this round")
        pool = existing_pools[0]
        if pool.design_job_id is None:
            # No design job associated with this pool, raise an error
            raise ValueError(f"Pool with name '{pool_name}' already exists in this round")
        design_job = db.get(DesignJob, id=pool.design_job_id)
        try:
            before = design_job.workflow.prepare_params(workdir=scheduler.workdir)
            after = workflow.prepare_params(workdir=scheduler.workdir)
        except Exception as e:
            traceback.print_exc()
            print(f"Error comparing existing pool workflow params: {e}")
            raise ValueError(
                f"Pool with name '{pool_name}' already exists in this round and could not compare parameters"
            )

        if before != after:
            print("Differences:")
            for k in set(before).union(after):
                if before.get(k) != after.get(k):
                    print(f"  {k}: {before.get(k)} -> {after.get(k)}")
            raise ValueError(
                f"Please choose a different pool name. Pool with name '{pool_name}' was already submitted in this round with different parameters."
            )
        print("Pool with same name and params already exists in this round, returning existing pool")
        return design_job, pool

    if config.props.read_only:
        raise RuntimeError("Cannot submit design workflow: OVO server is in read-only mode")

    # Create a deep copy of the workflow object to avoid errors
    # when users modify workflow params in place and resubmit
    workflow = deepcopy(workflow)

    workflow.validate()

    username = get_username()

    job_id = scheduler.submit(
        pipeline_name=pipeline_name or workflow.get_pipeline_name(),
        params=workflow.prepare_params(workdir=scheduler.workdir),
    )

    design_job = DesignJob(
        workflow=workflow,
        job_id=job_id,
        scheduler_key=scheduler_key,
        author=username,
    )

    pool = Pool(
        id=Pool.generate_id(),
        author=username,
        round_id=round_id,
        name=pool_name,
        description=pool_description,
        design_job_id=design_job.id,
        processed=False,
    )

    # TODO cancel job if this fails
    db.save_all([design_job, pool])

    return design_job, pool


def get_log(design_job: DesignJob, tail: int = None) -> str:
    """Get the log of a design job from the scheduler."""
    assert isinstance(design_job, DesignJob), f"Expected DesignJob, got {type(design_job).__name__}"
    scheduler = get_scheduler(design_job.scheduler_key)
    log = scheduler.get_log(design_job.job_id)
    if tail is not None:
        log_lines = log.splitlines()
        log = "\n".join(log_lines[-tail:])
    return log


def process_results(design_job: DesignJob, callback: Callable = None, wait=True) -> Pool:
    """Process the results of a design job and return the pool object.

    This downloads/copies the workflow results into Storage
    and saves the Design and DescriptorValue objects to the database."""
    assert isinstance(design_job, DesignJob), f"Expected DesignJob, got {type(design_job).__name__}"

    pool = db.get(Pool, design_job_id=design_job.id)
    if pool.processed:
        # Pool already processed
        print("Pool already processed, returning existing pool")
        return pool
    scheduler = get_scheduler(design_job.scheduler_key)
    if scheduler.get_result(design_job.job_id) is None:
        # Job is still running
        if wait:
            print(f"Waiting for job {design_job.job_id} to finish...")
            scheduler.wait(design_job.job_id)
        else:
            raise ValueError(
                f"Job {design_job.job_id} has not finished yet, try again later or use scheduler.wait(design_job.job_id)"
            )
    # Job should be successful or failed now, update job status
    update_job_status(design_job)
    if design_job.job_result is False:
        # Print job log and raise error
        print(scheduler.get_log(design_job.job_id))
        print(scheduler.get_failed_message(design_job.job_id), file=sys.stderr)
        print()
        raise ValueError(f"Job {design_job.job_id} has failed")
    # Process results
    print("Job finished, processing pool results...")
    objects = design_job.workflow.process_results(design_job, callback=callback)
    assert isinstance(objects, list), f"Expected list from process_results(), got {type(objects).__name__}: {objects}"
    # Mark pool as processed
    pool.processed = True
    flag_modified(design_job, "workflow")
    flag_modified(design_job, "warnings")
    # Atomically save all objects
    db.save_all(objects + [pool, design_job])
    return pool


def update_acceptance_thresholds(pools: list[Pool], acceptance_thresholds: dict[str, Threshold]):
    """Update the accepted designs in a pool based on the given acceptance thresholds.

    Save the Design objects and the DesignJob.workflow.acceptance_thresholds to the DB.
    """
    for pool in pools:
        assert isinstance(pool, Pool), f"Expected collection of Pool objects, got {pool} ({type(pool).__name__})"
        assert pool.design_job_id is not None, f"Pool must have a design job to update acceptance thresholds: {pool.id}"
    all_design_ids = db.select_values(Design, "id", pool_id__in=[p.id for p in pools])
    design_jobs = db.select(DesignJob, id__in=[p.design_job_id for p in pools])
    new_accepted_design_ids, num_accepted_by_descriptor = filter_designs_by_thresholds(
        all_design_ids=all_design_ids,
        thresholds=acceptance_thresholds,
        values={
            descriptor_key: db.select_descriptor_values(descriptor_key, design_ids=all_design_ids).to_dict()
            for descriptor_key, threshold in acceptance_thresholds.items()
            if threshold.enabled
        },
    )
    # Save the acceptance thresholds to the DesignJob
    for job in design_jobs:
        job.workflow.acceptance_thresholds = acceptance_thresholds
        # tell sqlalchemy that we modified the workflow object - with dataclasses this is not auto-detected
        flag_modified(job, "workflow")
    db.save_all(design_jobs)
    # Save accepted field of Design objects
    update_accepted_design_ids(pool_ids=[p.id for p in pools], accepted_design_ids=new_accepted_design_ids)


def update_accepted_design_ids(pool_ids: list[str], accepted_design_ids: list[str]):
    db.save_value(Design, "accepted", False, pool_id__in=pool_ids)
    db.save_value(Design, "accepted", True, id__in=accepted_design_ids)
    saved_ids = set(db.select_values(Design, "id", pool_id__in=pool_ids, accepted=True))
    assert saved_ids == set(accepted_design_ids), (
        f"Unexpected error marking designs as accepted "
        f"({saved_ids.difference(accepted_design_ids)} extra, "
        f"{set(accepted_design_ids).difference(saved_ids)} missing)"
    )


def set_designs_accepted(
    designs: list[Design], descriptor_values: list[DescriptorValue], thresholds: dict[str, Threshold]
):
    """Update the accepted field of designs based on the given thresholds (does not save to DB)"""
    # initialize dict of dicts (descriptor_key -> design_id -> value)
    values = {}
    for descriptor_value in descriptor_values:
        if descriptor_value.descriptor_key not in thresholds:
            continue
        if descriptor_value.descriptor_key not in values:
            values[descriptor_value.descriptor_key] = {}
        values[descriptor_value.descriptor_key][descriptor_value.design_id] = (
            float(descriptor_value.value) if descriptor_value.value is not None else None
        )
    # get list of accepted design ids based on thresholds
    accepted_design_ids, _ = filter_designs_by_thresholds(
        all_design_ids=[d.id for d in designs],
        thresholds=thresholds,
        values=values,
    )
    # update the accepted field of designs
    accepted_design_ids = set(accepted_design_ids)
    for design in designs:
        design.accepted = design.id in accepted_design_ids


def collect_storage_paths(download_fields: dict[str, tuple[Base, str]], design_ids: list) -> list[str]:
    """Collect storage paths from designs based on fields returned by DesignWorkflow.get_download_fields()."""
    storage_paths = []
    for label, (Model, field_name) in download_fields.items():
        if "." in field_name:
            # Allow nesting like rfdiffusion_params.input_pdb_path
            assert field_name.count(".") == 1, f"Only single-level relationships supported, got: {field_name}"
            field_name, subfield_name = field_name.split(".")
        else:
            # Regular field like structure_path
            subfield_name = None

        if issubclass(Model, Design):
            # Get field values from Design objects
            field_values = db.select_unique_values(Model, field_name, id__in=design_ids)
        elif issubclass(Model, DesignWorkflow):
            # Get field values from DesignWorkflow objects via the DesignJob associated with each Pool
            pool_ids = db.select_unique_values(Design, "pool_id", id__in=design_ids)
            pools = db.select(Pool, id__in=pool_ids)
            design_job_ids = [pool.design_job_id for pool in pools if pool.design_job_id]
            field_values = [
                getattr(workflow, field_name)
                for workflow in db.select_values(DesignJob, "workflow", id__in=design_job_ids)
                if isinstance(workflow, Model)
            ]
        elif issubclass(Model, DescriptorValue):
            field_values = db.select_unique_values(Model, "value", descriptor_key=field_name, design_id__in=design_ids)
        else:
            raise ValueError(
                f"Unsupported model {Model.__name__} for download, use a subclass of Design or DesignWorkflow"
            )

        for field_value in field_values:
            if not field_value:
                continue
            path = field_value if subfield_name is None else getattr(field_value, subfield_name)
            if not path:
                continue
            elif isinstance(path, str):
                storage_paths.append(path)
            elif isinstance(path, list):
                storage_paths.extend(path)
            else:
                raise ValueError(
                    f"Unexpected field type {type(path)} for {field_name}.{subfield_name}, expected str or list"
                )
    return storage_paths



from sqlalchemy.exc import NoResultFound

from ovo import db, get_username
from ovo.core.database.models import UserSettings


def get_or_create_user_settings(username: str = None) -> UserSettings:
    if username is None:
        username = get_username()
    try:
        user_settings = db.get(UserSettings, username=get_username())
    except NoResultFound:
        user_settings = UserSettings(username=username)
        db.save(user_settings)
    return user_settings
